from __future__ import annotations
import binascii
from typing import Dict, Union
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http2.result import HTTP2Result
from .action import GRPCAction


class GRPCResult(HTTP2Result):

    def __init__(self, action: GRPCAction, error: Exception = None) -> None: 
        super(GRPCResult, self).__init__(action, error)
        self.type = RequestTypes.GRPC

    @property
    def data(self):
        wire_msg = binascii.b2a_hex(self.body)

        message_length = wire_msg[4:10]
        msg = wire_msg[10:10+int(message_length, 16)*2]

        return binascii.a2b_hex(msg)

    @data.setter
    def data(self, value):
        self.body = value

    def to_protobuf(self, protobuf):
        protobuf.ParseFromString(self.data)
        return protobuf



import asyncio
from concurrent.futures import ThreadPoolExecutor
import inspect
from typing import Dict, Generic, TypeVar

import psutil
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.registrar import plugin_registrar


T = TypeVar('T')


class ReporterPlugin(Generic[T], Plugin):
    type=PluginType.REPORTER
    config: T = None
    name: str=None

    def __init__(self, config: T) -> None:
        
        super(
            ReporterPlugin,
            self
        ).__init__()

        self.loop = asyncio.get_event_loop()
        self.executor = ThreadPoolExecutor(
             max_workers=psutil.cpu_count(logical=False)
        )
        self.name = self.name

        self.hooks: Dict[PluginHooks, PluginHook] = {}
    
        self.config = config

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook

        connect_hook = self.hooks.get(PluginHooks.ON_REPORTER_CONNECT)
        close_hook = self.hooks.get(PluginHooks.ON_REPORTER_CLOSE)
        
        submit_events_hook = self.hooks.get(PluginHooks.ON_PROCESS_EVENTS)
        submit_shared_stats_hook = self.hooks.get(PluginHooks.ON_PROCESS_SHARED_STATS)
        submit_metrics_hook = self.hooks.get(PluginHooks.ON_PROCESS_METRICS)
        submit_custom_stats_hook = self.hooks.get(PluginHooks.ON_PROCESS_CUSTOM_STATS)
        submit_errors_hook = self.hooks.get(PluginHooks.ON_PROCESS_ERRORS)
        
        self.connect = connect_hook.call
        self.close = close_hook.call

        self.submit_events = submit_events_hook.call
        self.submit_common = submit_shared_stats_hook.call
        self.submit_metrics = submit_metrics_hook.call
        self.submit_custom = submit_custom_stats_hook.call
        self.submit_errors = submit_errors_hook.call



from enum import Enum
from hedra.monitoring import (
    MemoryMonitor,
    CPUMonitor
)
from hedra.reporting.metric.metric_types import MetricType
from hedra.reporting.tags import Tag
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat
)
from typing import (
    Dict,
    List,
    Union
)


MemoryMonitorGroup = Dict[str, MemoryMonitor]

CPUMonitorGroup = Dict[str, CPUMonitor]

MonitorGroup = Dict[str, Union[CPUMonitorGroup, MemoryMonitorGroup]]

StageSystemMetricsGroup = Dict[str, Dict[str, List[Union[int, float]]]]


class SystemMetricGroupType(Enum):
    CPU='cpu'
    MEMORY='memory'


class SessionMetricsCollection(BaseModel):
    name: StrictStr
    group: StrictStr
    mean: Union[StrictInt, StrictFloat]
    median: Union[StrictInt, StrictFloat]
    max: Union[StrictInt, StrictFloat]
    min: Union[StrictInt, StrictFloat]
    stdev: Union[StrictInt, StrictFloat]
    variance: Union[StrictInt, StrictFloat]
    quantile_10th: Union[StrictInt, StrictFloat]
    quantile_20th: Union[StrictInt, StrictFloat]
    quantile_25th: Union[StrictInt, StrictFloat]
    quantile_30th: Union[StrictInt, StrictFloat]
    quantile_40th: Union[StrictInt, StrictFloat]
    quantile_50th: Union[StrictInt, StrictFloat]
    quantile_60th: Union[StrictInt, StrictFloat]
    quantile_70th: Union[StrictInt, StrictFloat]
    quantile_75th: Union[StrictInt, StrictFloat]
    quantile_80th: Union[StrictInt, StrictFloat]
    quantile_90th: Union[StrictInt, StrictFloat]
    quantile_95th: Union[StrictInt, StrictFloat]
    quantile_99th: Union[StrictInt, StrictFloat]

    @property
    def stats(self) -> Dict[str, Union[int, float]]:
        return {
            'mean': self.mean,
            'median': self.median,
            'max': self.max,
            'min': self.min,
            'stdev': self.stdev,
            'variance': self.variance,
            'quantile_10th': self.quantile_10th,
            'quantile_20th': self.quantile_20th,
            'quantile_25th': self.quantile_25th,
            'quantile_30th': self.quantile_30th,
            'quantile_40th': self.quantile_40th,
            'quantile_50th': self.quantile_50th,
            'quantile_60th': self.quantile_60th,
            'quantile_70th': self.quantile_70th,
            'quantile_75th': self.quantile_75th,
            'quantile_80th': self.quantile_80th,
            'quantile_90th': self.quantile_90th,
            'quantile_95th': self.quantile_95th,
            'quantile_99th': self.quantile_99th
        }

    @property
    def record(self) -> Dict[str, Union[str, bool, int, float]]:
        return {
            'name': self.name,
            'group': self.group,
            **self.stats
        }
    
    @property
    def types_map(self):
        return {
            'mean': MetricType.SAMPLE,
            'median': MetricType.SAMPLE,
            'max': MetricType.SAMPLE,
            'min': MetricType.SAMPLE,
            'stdev': MetricType.SAMPLE,
            'variance': MetricType.SAMPLE,
            'quantile_10th': MetricType.SAMPLE,
            'quantile_20th': MetricType.SAMPLE,
            'quantile_25th': MetricType.SAMPLE,
            'quantile_30th': MetricType.SAMPLE,
            'quantile_40th': MetricType.SAMPLE,
            'quantile_50th': MetricType.SAMPLE,
            'quantile_60th': MetricType.SAMPLE,
            'quantile_70th': MetricType.SAMPLE,
            'quantile_75th': MetricType.SAMPLE,
            'quantile_80th': MetricType.SAMPLE,
            'quantile_90th': MetricType.SAMPLE,
            'quantile_95th': MetricType.SAMPLE,
            'quantile_99th': MetricType.SAMPLE,
        }
    
    @property
    def tags(self):
        tag_fields = {
            'name': self.name,
            'group': self.group
        }

        return [
            Tag(
                tag_field_name,
                tag_field_value
            ) for tag_field_name, tag_field_value in tag_fields.items()
        ]


class SystemMetricsCollection(BaseModel):
    stage: StrictStr
    name: StrictStr
    group: StrictStr
    mean: Union[StrictInt, StrictFloat]
    median: Union[StrictInt, StrictFloat]
    max: Union[StrictInt, StrictFloat]
    min: Union[StrictInt, StrictFloat]
    stdev: Union[StrictInt, StrictFloat]
    variance: Union[StrictInt, StrictFloat]
    quantile_10th: Union[StrictInt, StrictFloat]
    quantile_20th: Union[StrictInt, StrictFloat]
    quantile_25th: Union[StrictInt, StrictFloat]
    quantile_30th: Union[StrictInt, StrictFloat]
    quantile_40th: Union[StrictInt, StrictFloat]
    quantile_50th: Union[StrictInt, StrictFloat]
    quantile_60th: Union[StrictInt, StrictFloat]
    quantile_70th: Union[StrictInt, StrictFloat]
    quantile_75th: Union[StrictInt, StrictFloat]
    quantile_80th: Union[StrictInt, StrictFloat]
    quantile_90th: Union[StrictInt, StrictFloat]
    quantile_95th: Union[StrictInt, StrictFloat]
    quantile_99th: Union[StrictInt, StrictFloat]

    @property
    def stats(self) -> Dict[str, Union[int, float]]:
        return {
            'mean': self.mean,
            'median': self.median,
            'max': self.max,
            'min': self.min,
            'stdev': self.stdev,
            'variance': self.variance,
            'quantile_10th': self.quantile_10th,
            'quantile_20th': self.quantile_20th,
            'quantile_25th': self.quantile_25th,
            'quantile_30th': self.quantile_30th,
            'quantile_40th': self.quantile_40th,
            'quantile_50th': self.quantile_50th,
            'quantile_60th': self.quantile_60th,
            'quantile_70th': self.quantile_70th,
            'quantile_75th': self.quantile_75th,
            'quantile_80th': self.quantile_80th,
            'quantile_90th': self.quantile_90th,
            'quantile_95th': self.quantile_95th,
            'quantile_99th': self.quantile_99th
        }

    @property
    def record(self) -> Dict[str, Union[str, bool, int, float]]:
        return {
            'stage': self.stage,
            'name': self.name,
            'group': self.group,
            **self.stats
        }
    
    @property
    def types_map(self):
        return {
            'mean': MetricType.SAMPLE,
            'median': MetricType.SAMPLE,
            'max': MetricType.SAMPLE,
            'min': MetricType.SAMPLE,
            'stdev': MetricType.SAMPLE,
            'variance': MetricType.SAMPLE,
            'quantile_10th': MetricType.SAMPLE,
            'quantile_20th': MetricType.SAMPLE,
            'quantile_25th': MetricType.SAMPLE,
            'quantile_30th': MetricType.SAMPLE,
            'quantile_40th': MetricType.SAMPLE,
            'quantile_50th': MetricType.SAMPLE,
            'quantile_60th': MetricType.SAMPLE,
            'quantile_70th': MetricType.SAMPLE,
            'quantile_75th': MetricType.SAMPLE,
            'quantile_80th': MetricType.SAMPLE,
            'quantile_90th': MetricType.SAMPLE,
            'quantile_95th': MetricType.SAMPLE,
            'quantile_99th': MetricType.SAMPLE,
        }
    
    @property
    def tags(self):
        tag_fields = {
            'stage': self.stage,
            'name': self.name,
            'group': self.group
        }

        return [
            Tag(
                tag_field_name,
                tag_field_value
            ) for tag_field_name, tag_field_value in tag_fields.items()
        ]


from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from typing import (
    Dict, 
    Union, 
    Any, 
    List,
    Callable,
    Coroutine
)

from .aws_lambda.aws_lambda_connector import AWSLambdaConnector
from .aws_lambda.aws_lambda_connector_config import AWSLambdaConnectorConfig
from .bigtable.bigtable_connector import BigTableConnector
from .bigtable.bigtable_connector_config import BigTableConnectorConfig
from .cassandra.cassandra_connector import CassandraConnector
from .cassandra.cassandra_connector_config import CassandraConnectorConfig
from .common.connector_type import ConnectorType
from .cosmosdb.cosmos_connector import CosmosDBConnector
from .cosmosdb.cosmos_connector_config import CosmosDBConnectorConfig
from .csv.csv_connector import CSVConnector
from .csv.csv_connector_config import CSVConnectorConfig
from .google_cloud_storage.google_cloud_storage_connector import GoogleCloudStorageConnector
from .google_cloud_storage.google_cloud_storage_connector_config import GoogleCloudStorageConnectorConfig
from .har.har_connector import HARConnector
from .har.har_connector_config import HARConnectorConfig
from .json.json_connector import JSONConnector
from .json.json_connector_config import JSONConnectorConfig
from .kafka.kafka_connector import KafkaConnector
from .kafka.kafka_connector_config import KafkaConnectorConfig
from .mongodb.mongodb_connector import MongoDBConnector
from .mongodb.mongodb_connector_config import MongoDBConnectorConfig
from .mysql.mysql_connector import MySQLConnector
from .mysql.mysql_connector_config import MySQLConnectorConfig
from .postgres.postgres_connector import PostgresConnection
from .postgres.postgres_connector_config import PostgresConnectorConfig
from .redis.redis_connector import RedisConnector
from .redis.redis_connector_config import RedisConnectorConfig
from .s3.s3_connector import S3Connector
from .s3.s3_connector_config import S3ConnectorConfig
from .snowflake.snowflake_connector import SnowflakeConnector
from .snowflake.snowflake_connector_config import SnowflakeConnectorConfig
from .sqlite.sqlite_connector import SQLiteConnector
from .sqlite.sqlite_connector_config import SQLiteConnectorConfig
from .xml.xml_connector import XMLConnector
from .xml.xml_connector_config import XMLConnectorConfig


ConnectorConfig = Union[
    AWSLambdaConnectorConfig,
    BigTableConnectorConfig,
    CassandraConnectorConfig,
    CosmosDBConnectorConfig,
    CSVConnectorConfig,
    GoogleCloudStorageConnectorConfig,
    HARConnectorConfig,
    JSONConnectorConfig,
    KafkaConnectorConfig,
    MongoDBConnectorConfig,
    MySQLConnectorConfig,
    PostgresConnectorConfig,
    RedisConnectorConfig,
    S3ConnectorConfig,
    SnowflakeConnectorConfig,
    SQLiteConnectorConfig,
    XMLConnectorConfig
]


class Connector:

    def __init__(
        self, 
        stage: str,
        connector_config: ConnectorConfig,
        parser_config: Config
    ) -> None:
        self._connectors: Dict[
            ConnectorType,
            Callable[
                [
                    Union[
                        AWSLambdaConnectorConfig,
                        BigTableConnectorConfig,
                        CassandraConnectorConfig,
                        CosmosDBConnectorConfig,
                        CSVConnectorConfig,
                        GoogleCloudStorageConnectorConfig,
                        HARConnectorConfig,
                        JSONConnectorConfig,
                        KafkaConnectorConfig,
                        MongoDBConnectorConfig,
                        MySQLConnectorConfig,
                        PostgresConnectorConfig,
                        RedisConnectorConfig,
                        S3ConnectorConfig,
                        SnowflakeConnectorConfig,
                        SQLiteConnectorConfig,
                        XMLConnectorConfig
                    ],
                    str,
                    Config
                ],
                Union[
                    AWSLambdaConnector,
                    BigTableConnector
                ]
            ]
        ] = {
            ConnectorType.AWSLambda: lambda config, stage, parser_config: AWSLambdaConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.BigTable: lambda config, stage, parser_config: BigTableConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.Cassandra: lambda config, stage, parser_config: CassandraConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.CosmosDB: lambda config, stage, parser_config: CosmosDBConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.CSV: lambda config, stage, parser_config: CSVConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.GCS: lambda config, stage, parser_config: GoogleCloudStorageConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.HAR: lambda config, stage, parser_config: HARConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.JSON: lambda config, stage, parser_config: JSONConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.Kafka: lambda config, stage, parser_config: KafkaConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.MongoDB: lambda config, stage, parser_config: MongoDBConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.MySQL: lambda config, stage, parser_config: MySQLConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.Postgres: lambda config, stage, parser_config: PostgresConnection(
                config,
                stage,
                parser_config
            ),
            ConnectorType.Redis: lambda config, stage, parser_config: RedisConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.S3: lambda config, stage, parser_config: S3Connector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.Snowflake: lambda config, stage, parser_config: SnowflakeConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.SQLite: lambda config, stage, parser_config: SQLiteConnector(
                config,
                stage,
                parser_config
            ),
            ConnectorType.XML: lambda config, stage, parser_config: XMLConnector(
                config,
                stage,
                parser_config
            )
        }   

        self.selected = self._connectors.get(
            connector_config.connector_type
        )(
            connector_config,
            stage,
            parser_config
        )

        self.stage = stage
        self.parser_config = parser_config
        self.connected = False

    async def connect(self):
        await self.selected.connect()
        self.connected = True

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        return await self.selected.load_execute_stage_summary(
            options=options
        )

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        return await self.selected.load_actions(
            options=options
        )
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        return await self.selected.load_results(
            options=options
        )
    
    async def load_data(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, Any]:
        return await self.load_data(
            options=options
        )
    
    async def close(self):
        self.connected = False
        return await self.selected.close()


from hedra.core.experiments.variant import Variant
from scipy.optimize import least_squares
from typing import Dict, List, Tuple, Union
from .base_algorithm import BaseAlgorithm


class LeastSquaresOptimizer(BaseAlgorithm):

    def __init__(
        self, 
        config: Dict[str, Union[List[Tuple[Union[int, float]]], int]]
    ) -> None:
        super().__init__(config)
        self.initial_distribution: List[float] = []

    def generate_initial_distribution(self, variant: Variant):
        self.initial_distribution = variant.distribution.generate(self.stage_config.batch_size)

    def optimize(self, func):
        return least_squares(
            func,
            self.initial_distribution,
            diff_step=[
                0.5 for _ in self.distribution
            ],
            max_nfev=self.max_iter
        )


import importlib
import ntpath
import sys
import asyncio
import inspect
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List
from hedra.logging import HedraLogger
from hedra.core.hooks.types.event.event import Event, EventHook
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.hooks.types.base.hook import Hook
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.plugin import Plugin
from .stage import Stage


def import_from_path(path: str) -> Any:
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)
    
    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)

    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)

    return module


def import_stages(path: str) -> Dict[str, Stage]:
    
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())

    discovered = {}
    for name, stage_candidate in inspect.getmembers(module):
        if inspect.isclass(
            stage_candidate
        ) and issubclass(
            stage_candidate, 
            Stage
        ) and stage_candidate not in direct_decendants:
            discovered[name] = stage_candidate

    return discovered


def import_plugins(path: str) -> Dict[PluginType, Dict[str, Plugin]]:
    module = import_from_path(path)
    direct_decendants = list({cls.__name__: cls for cls in Plugin.__subclasses__()}.values())

    plugins_by_type = defaultdict(dict)
    for name, plugin_candidate in inspect.getmembers(module):
        if inspect.isclass(
            plugin_candidate
        ) and issubclass(
            plugin_candidate, 
            Plugin
        ) and plugin_candidate not in direct_decendants:
            plugins_by_type[plugin_candidate.type][plugin_candidate.name] = plugin_candidate

    return plugins_by_type


def set_stage_hooks(stage: Stage, generated_hooks: Dict[str, Hook]) -> Stage:
    methods = inspect.getmembers(stage, predicate=inspect.ismethod) 

    for _, method in methods:
        method_name = method.__qualname__
        hook_set: List[Hook] = registrar.all.get(method_name, [])

        for hook in hook_set:

            if generated_hooks.get(hook) is None:
                hook = hook.copy()
                hook._call = hook._call.__get__(stage, stage.__class__)
                setattr(stage, hook.shortname, hook._call)

                if inspect.ismethod(hook.call) is False:
                    hook.call = hook.call.__get__(stage, stage.__class__)
                    setattr(stage, hook.shortname, hook.call)

                generated_hooks[hook] = 'created'
                hook.stage = stage.name
                
                hook.stage_instance: Stage = stage

                hook.name = f'{hook.stage}.{hook.shortname}'
                stage.hooks[hook.hook_type].append(hook)
            

            elif generated_hooks.get(hook) == 'created':

                copied_hook = hook.copy()

                stage_config: Dict[str, Any]  = stage.to_copy_dict()
                copied_stage = type(stage)()

                for copied_attribute_name, copied_attribute_value in stage_config.items():
                    if inspect.ismethod(copied_attribute_value) is False:
                        setattr(copied_stage, copied_attribute_name, copied_attribute_value)
                        
                copied_hook.stage = stage.name
                copied_hook.stage_instance: Stage = copied_stage
                copied_hook.name = f'{stage.name}.{hook.shortname}'
                copied_hook._call = method

                copied_hook.name = f'{copied_hook.stage}.{copied_hook.shortname}'

                copied_hook._call = copied_hook._call.__get__(stage, stage.__class__)
                setattr(stage, copied_hook.shortname, copied_hook._call)

                if inspect.ismethod(copied_hook.call) is False:
                    copied_hook.call = copied_hook.call.__get__(stage, stage.__class__)
                    setattr(stage, copied_hook.shortname, copied_hook.call)

                if copied_hook not in stage.hooks[hook.hook_type]:
                    stage.hooks[hook.hook_type].append(copied_hook)

    return stage



from typing import Union
from scipy.stats import triang
from .base import BaseDistribution


class TriangularDistribution(BaseDistribution):

    def __init__(
        self,
        size: int,
        alpha: float=0.25,
        beta: float=0.5,
        center: Union[int, float]=0.5,
        randomness: Union[int, float]=0.25
    ):
        super().__init__(
            size=size,
            center=center,
            randomness=randomness,
            frozen_distribution=triang(
                alpha,
                beta,
                loc=center,
                scale=randomness
            )
        )



from hedra.distributed.env import Env
from hedra.distributed.models.http import (
    Limit,
    Request
)
from pydantic import IPvAnyAddress
from typing import (
    Optional,
    Callable,
    Dict,
    Union
)
from .limiters import (
    AdaptiveRateLimiter,
    CPUAdaptiveLimiter,
    LeakyBucketLimiter,
    ResourceAdaptiveLimiter,
    SlidingWindowLimiter,
    TokenBucketLimiter
)

class Limiter:

    def __init__(
        self,
        env: Env
    ) -> None:
        
        self._limiter: Union[
            Union[
                AdaptiveRateLimiter,
                CPUAdaptiveLimiter,
                LeakyBucketLimiter,
                ResourceAdaptiveLimiter,
                SlidingWindowLimiter,
                TokenBucketLimiter
            ], 
            None
        ] = None

        self._default_limit = Limit(
            max_requests=env.MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS,
            request_period=env.MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD,
            reject_requests=env.MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT,
            cpu_limit=env.MERCURY_SYNC_HTTP_CPU_LIMIT,
            memory_limit=env.MERCURY_SYNC_HTTP_MEMORY_LIMIT
        )
        
        self._rate_limit_strategy = env.MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY
        self._default_limiter_type = env.MERCURY_SYNC_HTTP_RATE_LIMITER_TYPE
        
        self._rate_limiter_types: Dict[
            str,
            Callable[
                [Limit],
                Union[
                    AdaptiveRateLimiter,
                    CPUAdaptiveLimiter,
                    LeakyBucketLimiter,
                    ResourceAdaptiveLimiter,
                    SlidingWindowLimiter,
                    TokenBucketLimiter
                ]
            ]
        ] = {
            "adaptive": AdaptiveRateLimiter,
            "cpu-adaptive": CPUAdaptiveLimiter,
            "leaky-bucket": LeakyBucketLimiter,
            "rate-adaptive": ResourceAdaptiveLimiter,
            "sliding-window":  SlidingWindowLimiter,
            "token-bucket":  TokenBucketLimiter,
        }

        self._rate_limit_period = env.MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD

        self._rate_limiters: Dict[
            str, 
            Union[
                AdaptiveRateLimiter,
                CPUAdaptiveLimiter,
                LeakyBucketLimiter,
                SlidingWindowLimiter,
                TokenBucketLimiter
            ]
        ] = {}

    async def limit(
        self,
        ip_address: IPvAnyAddress,
        request: Request,
        limit: Optional[Limit]=None
    ):
        limit_key: Union[str, None] = None

        if self._rate_limit_strategy == 'ip':

            if limit is None:
                limit = self._default_limit

            limit_key = limit.get_key(
                request,
                ip_address,
                default=ip_address
            )

        elif self._rate_limit_strategy == 'endpoint' and limit:

            if limit is None:
                limit = self._default_limit

            limit_key = limit.get_key(
                request,
                ip_address,
                default=request.path
            )
        
        elif self._rate_limit_strategy == 'global':

            limit_key = self._default_limit.get_key(
                request,
                ip_address,
                default='default'
            )

            limit = self._default_limit

        elif self._rate_limit_strategy == "ip-endpoint" and limit:

            if limit is None:
                limit = self._default_limit

            limit_key = limit.get_key(
                request,
                ip_address,
                default=f'{request.path}_{ip_address}'
            )

        elif limit:

            limit_key = limit.get_key(
                request,
                ip_address
            )

        if limit_key and limit.matches(request, ip_address):
            return await self._check_limiter(
                limit_key,
                limit
            )

        return False 

    async def _check_limiter(
        self,
        limiter_key: str,
        limit: Limit
    ):
        
        limiter = self._rate_limiters.get(limiter_key)

        rate_limiter_type = limit.limiter_type
        if rate_limiter_type is None:
            rate_limiter_type = self._default_limiter_type

        if limiter is None:

            limiter = self._rate_limiter_types.get(rate_limiter_type)(limit)

            self._rate_limiters[limiter_key] = limiter
            

        return await limiter.acquire()
    
    async def close(self):
        for limiter in self._rate_limiters.values():
            if isinstance(limiter, CPUAdaptiveLimiter):
                await limiter.close()


import functools
from typing import Dict, Union, Tuple
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.registrar import registrar
from .validator import TaskHookValidator


@registrar(HookType.TASK)
def task(
    *names: Tuple[str, ...],
    weight: int=1, 
    order: int=1, 
    skip: bool=False,
    metadata: Dict[str, Union[str, int]]={}
):
    
    TaskHookValidator(
        names=names,
        weight=weight,
        order=order,
        skip=skip,
        metadata=metadata
    )

    def wrapper(func) -> Hook:

        @functools.wraps(func)
        def decorator(*args, **kwargs):

            return func(*args, **kwargs)
                
        return decorator

    return wrapper


from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.hooks import (
    action
)


class ExecuteHTTPStage(Execute):

    @action()
    async def http_get(self):
        return await self.client.http.get('https://<url_here>')


import math
import random
from typing import List, Dict, Union, Iterable
from hedra.versioning.flags.types.unstable.flag import unstable
from .variant import Variant


@unstable
class Experiment:

    def __init__(
        self,
        experiment_name: str,
        participants: List[Variant],
        random: bool=True
    ) -> None:
        self.experiment_name = experiment_name
        self.participants: Dict[str, Variant] = {
            participant.stage_name: participant for participant in participants
        }
        self.source_batch_size: int = 0
        self.random: bool = random
        self.participants_count = len(participants)

    def __iter__(self):
        for participant in self.participants.values():
            yield participant

    def assign_weights(self):

        total_weight = 1.0
        missing_weight_participants = []

        for participant_idx, participant in enumerate(self.participants.values()):
            
            if participant.weight:
                total_weight -= participant.weight

            elif self.random:
                weight  = round(random.uniform(0.1, 0.9), 2)

                if participant_idx < self.participants_count - 1:
                    participant.weight = weight
                    total_weight -= weight

                else:
                    participant.weight = round(total_weight, 2)

            else:
                missing_weight_participants.append(participant.stage_name)

            self.participants[participant.stage_name] = participant

        per_participant_weight = 0
        if len(missing_weight_participants) > 0:
            per_participant_weight = total_weight/len(missing_weight_participants)
            
        for missing_weight_participant in missing_weight_participants:
            self.participants[missing_weight_participant].weight = per_participant_weight
        

    def is_variant(self, stage_name: str) -> bool:
        variant = self.participants.get(stage_name)
        return variant is not None

    def get_variant_batch_size(self, stage_name: str) -> int:
        variant = self.participants.get(stage_name)

        if variant is None:
            raise Exception(
                f'Err. - requested variant {stage_name} is not a participant in experiment {self.experiment_name}.'
            )

        return math.ceil(variant.weight * self.source_batch_size)
    
    def get_variant(self, stage_name: str) -> Union[Variant, None]:
        if self.is_variant(stage_name):
            variant = self.participants.get(stage_name)

            return variant
        
    
    def calculate_distribution(
        self,
        stage_name: str,
        batch_size: int
    ) -> Union[List[float], None]:
        variant = self.get_variant(stage_name)

        if variant.distribution:
            return variant.distribution.generate(batch_size)

        


import asyncio
import socket
import signal
from asyncio.constants import SSL_HANDSHAKE_TIMEOUT
from ssl import SSLContext
from typing import Optional
from asyncio.sslproto import SSLProtocol
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from .protocol import TCPProtocol
from .tls_protocol import TLSProtocol
from hedra.core.engines.types.common.protocols.shared.constants import (
    _DEFAULT_LIMIT,
    _HTTP2_LIMIT
)


class TCPConnection:

    def __init__(self, factory_type: RequestTypes = RequestTypes.HTTP) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.transport = None
        self.factory_type = factory_type
        self._connection = None
        self.socket: socket.socket = None
        self._writer = None

    async def create(self, hostname=None, socket_config=None, *, limit=_DEFAULT_LIMIT, ssl=None):
        self.loop = asyncio.get_event_loop()

        family, type_, proto, _, address = socket_config

        socket_family = socket.AF_INET6 if family == 2 else socket.AF_INET

        self.socket = socket.socket(family=family, type=type_, proto=proto)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        
        await self.loop.run_in_executor(None, self.socket.connect, address)

        self.socket.setblocking(False)

        reader = Reader(limit=limit, loop=self.loop)
        reader_protocol = TCPProtocol(reader, loop=self.loop)

        if ssl is None:
            hostname = None

        self.transport, _ = await self.loop.create_connection(
            lambda: reader_protocol, 
            sock=self.socket,
            family=socket_family,
            server_hostname=hostname,
            ssl=ssl
        )

        self._writer = Writer(self.transport, reader_protocol, reader, self.loop)
        
        return reader, self._writer

    async def create_http2(self, hostname=None, socket_config=None, ssl: Optional[SSLContext] = None, ssl_timeout: int = SSL_HANDSHAKE_TIMEOUT):
        # this does the same as loop.open_connection(), but TLS upgrade is done
        # manually after connection be established.
        self.loop = asyncio.get_event_loop()

        family, type_, proto, _, address = socket_config
        
        socket_family = socket.AF_INET6 if family == 2 else socket.AF_INET

        self.socket = socket.socket(family=family, type=type_, proto=proto)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        await self.loop.run_in_executor(None, self.socket.connect, address)
        
        self.socket.setblocking(False)

        reader = Reader(limit=_HTTP2_LIMIT, loop=self.loop)

        protocol = TLSProtocol(reader, loop=self.loop)

        self.transport, _ = await self.loop.create_connection(
            lambda: protocol, 
            sock=self.socket,
            family=socket_family
        )
        
        ssl_protocol = SSLProtocol(
            self.loop, 
            protocol, 
            ssl, 
            None,
            False, 
            hostname,
            ssl_handshake_timeout=ssl_timeout,
            call_connection_made=False
        )

        # Pause early so that "ssl_protocol.data_received()" doesn't
        # have a chance to get called before "ssl_protocol.connection_made()".
        self.transport.pause_reading()

        self.transport.set_protocol(ssl_protocol)
        await self.loop.run_in_executor(None, ssl_protocol.connection_made, self.transport)
        self.transport.resume_reading()

        self.transport = ssl_protocol._app_transport

        reader = Reader(limit=_HTTP2_LIMIT, loop=self.loop)

        protocol.upgrade_reader(reader) # update reader
        protocol.connection_made(self.transport) # update transport

        self._writer = Writer(self.transport, ssl_protocol, reader, self.loop) # update writer

        return reader, self._writer

    async def close(self):

        try:
            self.transport._ssl_protocol.pause_writing()
            self.transport.close()

        except Exception:
            pass


import functools


def server():

    def wraps(func):

        func.server_only = True
        func.as_http = False

        @functools.wraps(func)
        def decorator(
            *args,
            **kwargs
        ):
            return func(*args, **kwargs)
        
        return decorator
    
    return wraps



from __future__ import annotations
from typing import (
    Dict,
    Iterator,
    List,
    Union
)
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.url import URL
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.action import HTTPAction



class HTTP3Action(HTTPAction):

    def __init__(
            self, 
            name: str, 
            url: str, 
            method: str = 'GET', 
            headers: Dict[str, str]={}, 
            data: Union[str, dict, Iterator, bytes, None] = None, 
            user: str = None, tags: 
            List[Dict[str, str]] = ..., redirects: int = 3
        ) -> None:

        super().__init__(
            name, 
            url, 
            method, 
            headers, 
            data, 
            user, 
            tags, 
            redirects
        )


        self.type = RequestTypes.HTTP3
        address_family, protocol = self.protocols[self.type]
        self.url = URL(url, family=address_family, protocol=protocol)


        self.redirects = redirects
        self.hooks: Hooks[HTTP3Action] = Hooks()
    
    def _setup_headers(self) -> Union[bytes, Dict[str, str]]:

        self.encoded_headers = [
            (b":method", self.method.encode()),
            (b":scheme", self.url.scheme.encode()),
            (b":authority", self.url.authority.encode()),
            (b":path", self.url.full.encode()),
            (b"user-agent", 'hedra/client'.encode()),
        ]
        
        self.encoded_headers.extend([
            (
                k.encode(), 
                v.encode()
            ) for (
                k, 
                v
            ) in self.headers.items()
        ])

    


from typing import List
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.idle.idle import Idle
from hedra.core.hooks.types.base.simple_context import SimpleContext


class IdleEdge(BaseEdge[Idle]):

    def __init__(self, source: Idle, destination: Stage) -> None:
        super(
            IdleEdge,
            self
        ).__init__(
            source,
            destination
        )

    async def transition(self):

        await self.source.run()

        if self.destination.context is None:
            self.destination.context = SimpleContext()

        self._update(self.destination)

        return None, self.destination.stage_type

    def _update(self, destination: Stage):
        for edge_name in self.history:

            history = self.history[edge_name]

            if self.next_history.get(edge_name) is None:
                self.next_history[edge_name] = {}

            self.next_history[edge_name].update({
                key: value for key, value  in history.items() if key in self.provides
            })


        self.next_history.update({
            (self.source.name, destination.name): {}
        })

    def split(self, edges: List[BaseEdge]) -> None:
        pass


from enum import Enum


class PluginHooks(Enum):
    ON_OPTIMIZE='ON_OPTIMIZE'
    ON_OPTIMIZER_GET_PARAMS='ON_OPTIMIZER_GET_PARAMS'
    ON_OPTIMIZER_UPDATE_PARAMS='ON_OPTIMIZER_UPDATE_PARAMS'
    ON_ENGINE_CONNECT='ON_ENGINE_CONNECT'
    ON_ENGINE_EXECUTE='ON_ENGINE_EXECUTE'
    ON_ENGINE_CLOSE='ON_ENGINE_CLOSE'
    ON_PERSONA_SETUP='ON_PERSONA_SETUP'
    ON_PERSONA_GENERATE='ON_PERSONA_GENERATE'
    ON_PERSONA_SHUTDOWN='ON_PERSONA_SHUTDOWN'
    ON_REPORTER_CONNECT='ON_REPORTER_CONNECT'
    ON_REPORTER_CLOSE='ON_REPORTER_CLOSE'
    ON_PROCESS_EVENTS='ON_PROCESS_EVENTS'
    ON_PROCESS_SHARED_STATS='ON_PROCESS_SHARED_STATS'
    ON_PROCESS_METRICS='ON_PROCESS_METRICS'
    ON_PROCESS_CUSTOM_STATS='ON_PROCESS_CUSTOM_STATS'
    ON_PROCESS_ERRORS='ON_PROCESS_ERRORS'
    ON_EXTENSION_EXECUTE='ON_EXTENSION_EXECUTE'
    ON_EXTENSION_PREPARE='ON_EXTENSION_PREPARE'
    CUSTOM='CUSTOM'


import time
import asyncio
import psutil
import uuid
import math
from typing import (
    Dict, 
    List, 
    Union,
    Optional
)
from hedra.logging import HedraLogger
from asyncio import Task
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.personas.batching.batch import Batch
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes
from hedra.core.personas.streaming import (
    Stream,
    StreamAnalytics
)
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.reporting.processed_result.results import results_types
from hedra.reporting.reporter import (
    Reporter,
    ReporterConfig
)



async def cancel_pending(pend: Task):
    try:
        if pend.done():
            pend.exception()

            return pend

        pend.cancel()
        await asyncio.sleep(0)
        if not pend.cancelled():
            await pend

        return pend
    
    except asyncio.CancelledError as cancelled_error:
        return cancelled_error

    except asyncio.TimeoutError as timeout_error:
        return timeout_error

    except asyncio.InvalidStateError as invalid_state:
        return invalid_state


class DefaultPersona:

    __slots__ = (
        'metadata_string',
        'persona_id',
        'logger',
        'type',
        'workers',
        'actions',
        '_hooks',
        'batch',
        '_stream',
        'total_time',
        'duration',
        'total_actions',
        'total_elapsed',
        'start',
        'end',
        'completed_actions',
        'pending_actions',
        'completed_time',
        'run_timer',
        'actions_count',
        'graceful_stop',
        'is_timed',
        '_stream_thread',
        '_loop',
        'current_action_idx',
        'optimized_params',
        '_executor',
        'stream',
        'streamed_analytics',
        'stream_reporter_configs',
        'stream_reporters',
        'stage_name',
        'optimization_active',
        'pending',
        'collect_analytics',
        'collection_interval',
        'bypass_cleanup',
        'cpu_monitor',
        'memory_monitor'
    )    

    def __init__(self, config: Config):
        self.persona_id = str(uuid.uuid4())
        self.logger = HedraLogger()
        self.logger.initialize()

        self.metadata_string: str = None

        self.type = PersonaTypes.DEFAULT
        self.workers = 1

        self.actions = []
        self._hooks: List[Union[ActionHook, TaskHook]] = {}
        self.batch = Batch(config)
        self._loop: asyncio.AbstractEventLoop = None

        self._stream = False
        self.optimization_active = False
        self.total_time = config.total_time
        self.duration = 0
        self.total_actions = 0
        self.total_elapsed = 0
        self.start = 0
        self.end = 0
        self.completed_actions = 0
        self.pending_actions = 0
        self.completed_time = 0
        self.run_timer = False
        self.actions_count = 0
        self.graceful_stop = config.graceful_stop
        self.stream: Stream = None 

        self.is_timed = True
        self._stream_thread = None

        self.current_action_idx = 0
        self.optimized_params = None
        self.streamed_analytics: StreamAnalytics = None
        self.stream_reporter_configs: List[ReporterConfig] = []
        self.stream_reporters: List[Reporter] = []
        self.stage_name: str = None
        self.collect_analytics: bool = False
        self.collection_interval: int = 1
        self.pending: List[asyncio.Task] = []
        self.bypass_cleanup: bool = False
        self.cpu_monitor = CPUMonitor()
        self.memory_monitor = MemoryMonitor()

    def setup(
            self, 
            hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]], 
            metadata_string: str
        ):

        self._setup(
            hooks, 
            metadata_string
        )

        self.actions_count = len(self._hooks)

    def _setup(
            self, 
            hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]], 
            metadata_string: str
        ):
        self.metadata_string = f'{metadata_string} Persona: {self.type.capitalize()}:{self.persona_id} - '

        actions_and_tasks = list([
            *hooks.get(HookType.ACTION),
            *hooks.get(HookType.TASK, [])
        ])

        self._hooks = [
            hook for hook in actions_and_tasks if not hook.skip
        ]

        for hook in self._hooks:
            if self.stage_name is None:
                self.stage_name = hook.stage
        
        for reporter_config in self.stream_reporter_configs:
            self.stream_reporters.append(
                Reporter(reporter_config)
            )

        self._stream = len(self.stream_reporters) > 0 and self.optimization_active is False

        if self._stream or self.collect_analytics:
            self.stream = Stream()

    async def set_concurrency(self, concurrency: int):
        for hook in self._hooks:
            await hook.session.set_pool(concurrency)

    async def execute(self):
        hooks = self._hooks
        hook_names = ', '.join([
            hook.name for hook in hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing {self.actions_count} Hooks: {hook_names}')

        total_time = self.total_time
        
        monitor_name = f'{self.stage_name}.persona'

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Executing for a total of - {total_time} - seconds')
        loop = asyncio.get_running_loop()
        if self._stream or self.collect_analytics:

            for reporter in self.stream_reporters:
                reporter.logger.filesystem.aio['hedra.reporting'].logger_enabled = False
                reporter.selected_reporter.logger.filesystem.aio['hedra.reporting'].logger_enabled = False
                await reporter.connect()

            
            await self.cpu_monitor.start_background_monitor(monitor_name)
            await self.memory_monitor.start_background_monitor(monitor_name)

            await self.start_stream()

            self.start = time.monotonic()
            completed, pending = await asyncio.wait([
                loop.create_task(
                    self.stream.execute_action(
                        hooks[action_idx]
                    )
                ) async for action_idx in self.generator(total_time)
            ], timeout=self.graceful_stop)

            self.end = time.monotonic()

            self.streamed_analytics = await self.stop_stream()

            await self.cpu_monitor.stop_background_monitor(monitor_name)
            await self.memory_monitor.stop_background_monitor(monitor_name)

            for reporter in self.stream_reporters:
                await reporter.close()

        else:

            await self.cpu_monitor.start_background_monitor(monitor_name)
            await self.memory_monitor.start_background_monitor(monitor_name)

            self.start = time.monotonic()
            completed, pending = await asyncio.wait([
                loop.create_task(
                    hooks[action_idx].session.execute_prepared_request(
                        hooks[action_idx].action
                    )
                ) async for action_idx in self.generator(total_time)
            ], timeout=self.graceful_stop)

            self.end = time.monotonic()

            await self.cpu_monitor.stop_background_monitor(monitor_name)
            await self.memory_monitor.stop_background_monitor(monitor_name)

        self.cpu_monitor.close()
        self.memory_monitor.close() 

        execution_elapsed = int(self.end - self.start)

        self.cpu_monitor.trim_monitor_samples(
            monitor_name,
            execution_elapsed
        )

        self.memory_monitor.trim_monitor_samples(
            monitor_name,
            execution_elapsed
        )

        self.pending_actions = len(pending)
        await self.logger.filesystem.aio['hedra.core'].debug(
            f'{self.metadata_string} - Execution completed with - {self.pending_actions} - actions left pending'
        )

        results = await asyncio.gather(*completed)

        cleanup_start = time.monotonic()

        await asyncio.gather(*[
            asyncio.create_task(
                cancel_pending(pend)
            ) for pend in pending
        ])

        cleanup_elapsed = time.monotonic() - cleanup_start
        await self.logger.filesystem.aio['hedra.core'].info(
            f'{self.metadata_string} - Cleanup completed - Resolved {self.pending_actions} pending actions in {round(cleanup_elapsed, 2)} seconds'
        )

        for hook in hooks:

            session_closed_start = time.monotonic()

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closing session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}')
            await hook.session.close()

            session_closed_elapsed = time.monotonic() - session_closed_start

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Closed session - {hook.session.session_id} - for Hook - {hook.name}:{hook.hook_id}. Took: {round(session_closed_elapsed, 2)} seconds')
        
        self.total_actions = len(set(results))
        self.total_elapsed = self.end - self.start
        self.optimized_params = None

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed execution')

        return results

    async def generator(self, total_time):
        elapsed = 0
        max_pool_size = math.ceil(self.batch.size * (psutil.cpu_count(logical=False)**2)/self.workers)
        action_idx = 0

        start = time.monotonic()
        while elapsed < total_time:
            yield action_idx
            await asyncio.sleep(0)
            elapsed = time.monotonic() - start

            if self._hooks[action_idx].session.active > max_pool_size:
                try:
                    max_wait = total_time - elapsed
                    await asyncio.wait_for(
                        self._hooks[action_idx].session.wait_for_active_threshold(),
                        timeout=max_wait
                    )
                except asyncio.TimeoutError:
                    pass
            
            action_idx = (action_idx+1)%self.actions_count

    async def start_stream(self):
        self._loop = asyncio.get_event_loop()
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Live Updates enabled')
        self._stream_thread = asyncio.create_task(
            self._run_stream()
        )

    async def stop_stream(self) -> StreamAnalytics:
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Live Updates stopped')
        self.run_timer = False
        return await self._stream_thread

    async def _run_stream(self) -> StreamAnalytics:
        self.run_timer = True

        stream_analytics = StreamAnalytics()
        
        stream_submission_tasks = []
        collection_stop_time = self.total_time - 1

        start = time.time()
        batch_start = time.time()

        while self.completed_time < collection_stop_time and self.run_timer:
            self.completed_time = time.time() - start
            await asyncio.sleep(self.collection_interval)

            batch_elapsed = time.time() - batch_start

            stream_analytics.add(
                self.stream,
                batch_elapsed
            )

            if self._stream:
                processed_results = [
                    results_types.get(
                        result.type
                    )(
                        self.stage_name,
                        result
                    ) for result in self.stream.completed
                ]

                for reporter in self.stream_reporters:
                    stream_submission_tasks.append(
                        asyncio.create_task(
                            reporter.submit_events(processed_results)
                        )
                    )

            batch_start = time.time()
            self.stream.completed = []

        self.stream.completed = []

        if self._stream:
            await asyncio.gather(*stream_submission_tasks)
   
        self.completed_time = time.time() - start

        return stream_analytics
        


from enum import Enum


class MetricType(Enum):
    COUNT='COUNT'
    RATE='RATE'
    DISTRIBUTION='DISTRIBUTION'
    SAMPLE='SAMPLE'



metric_type_map = {
    'count': MetricType.COUNT,
    'rate': MetricType.RATE,
    'distribution': MetricType.DISTRIBUTION,
    'sample': MetricType.SAMPLE
}


from hedra.distributed.middleware.base import (
    Middleware
)
from hedra.distributed.models.http import (
    Request,
    Response
)
from typing import Union, Optional, List, Literal, Tuple
from .cors_headers import CorsHeaders


class Cors(Middleware):

    def __init__(
        self,
        access_control_allow_origin: List[str]=None,
        access_control_allow_methods: List[
            Literal[
                "GET",
                "HEAD",
                "OPTIONS",
                "POST",
                "PUT",
                "PATCH",
                "DELETE",
                "TRACE"
            ]
        ]=None,
        access_control_expose_headers: Optional[List[str]]=None,
        access_control_max_age: Optional[Union[int, float]]=None,
        access_control_allow_credentials: Optional[bool]=None,
        access_control_allow_headers: Optional[List[str]]=None
    ) -> None:
         
        self._cors_config = CorsHeaders(
            access_control_allow_origin=access_control_allow_origin,
            access_control_expose_headers=access_control_expose_headers,
            access_control_max_age=access_control_max_age,
            access_control_allow_credentials=access_control_allow_credentials,
            access_control_allow_methods=access_control_allow_methods,
            access_control_allow_headers=access_control_allow_headers,
        )
        
        self.origins = self._cors_config.access_control_allow_origin
        self.cors_methods = self._cors_config.access_control_allow_methods
        self.cors_headers = self._cors_config.access_control_allow_headers
        self.allow_credentials = self._cors_config.access_control_allow_credentials

        self.allow_all_origins = '*' in self._cors_config.access_control_allow_origin

        allowed_headers = self._cors_config.access_control_allow_headers
        self.allow_all_headers = False

        if allowed_headers:
            self.allow_all_headers = '*' in allowed_headers

        self.simple_headers = self._cors_config.to_simple_headers()
        self.preflight_headers = self._cors_config.to_preflight_headers()
        self.preflight_explicit_allow_origin = not self.allow_all_origins or self.allow_credentials
        
        super().__init__(
            self.__class__.__name__,
            methods=['OPTIONS'],
            response_headers=self._cors_config.to_headers()
        )
        
    async def __run__(
        self, 
        request: Request,
        response: Optional[Response],
        status: Optional[int]
    ) -> Tuple[
        Tuple[Response, int],
        bool
    ]:

        headers = request.headers
        method = request.method

        origin = headers.get('origin')
        access_control_request_method = headers.get('access-control-request-method')
        access_control_request_headers = headers.get('access-control-request-headers')
        access_control_request_headers = headers.get("access-control-request-headers")

        if method == "OPTIONS" and access_control_request_method:

            response_headers = dict(self.preflight_headers)

            failures: List[str] = []

            if self.allow_all_origins is False and origin not in self.origins:
                failures.append("origin")

            elif self.preflight_explicit_allow_origin:
                response['Access-Control-Allow-Origin'] = origin

            if access_control_request_method not in self.cors_methods:
                failures.append("method")

            if self.allow_all_headers and access_control_request_headers is not None:
                response_headers["Access-Control-Allow-Headers"] = access_control_request_headers
            
            elif access_control_request_headers:

                for header in access_control_request_headers.split(
                    ','
                ):
                    if header.lower().strip() not in self.cors_headers:
                        failures.append("headers")
                        break
                
            if len(failures) > 0:

                failures_message = ', '.join(failures)

                return (
                    Response(
                        request.path,
                        request.method,
                        headers=response_headers,
                        data=f"Disallowed CORS {failures_message}"
                    ),
                    401
                ), False
            
            if response and status:
                response.headers.update(response_headers)

                return (
                    response, 
                    status
                ), False
            
            return (
                Response(
                    request.path,
                    request.method,
                    headers=response_headers,
                    data=None
                ),
                204
            ), False
        
        response_headers = dict(self.simple_headers)

        has_cookie = headers.get('cookie')
        if self.allow_all_origins and has_cookie:
            response_headers['access-control-allow-origin'] = origin

        elif origin in self.origins:
            response_headers['access-control-allow-origin'] = origin
        
        if response and status:
            response.headers.update(response_headers)

            return (
                response,
                status
            ), True

        return (
            Response(
                request.path,
                request.method,
                headers=response_headers,
                data=None
            ), 
            200
        ), True


class BooleanConfigOption:
    """
    Descriptor for handling a boolean config option.  This will block
    attempts to set boolean config options to non-bools.
    """
    def __init__(self, name):
        self.name = name
        self.attr_name = '_%s' % self.name

    def __get__(self, instance, owner):
        return getattr(instance, self.attr_name)

    def __set__(self, instance, value):
        if not isinstance(value, bool):
            raise ValueError("%s must be a bool" % self.name)
        setattr(instance, self.attr_name, value)


import asyncio
import math
import multiprocessing
import psutil
import warnings
from multiprocessing import active_children
from types import FunctionType
from concurrent.futures.process import BrokenProcessPool
from concurrent.futures import ProcessPoolExecutor
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from typing import (
    Any, 
    List, 
    Tuple, 
    Dict
)
from .synchronization import BatchedSemaphore
from .stage_priority import StagePriority


class BatchExecutor:

    def __init__(
        self, 
        max_workers: int = psutil.cpu_count(logical=False),
        start_method: str='spawn'
    ) -> None:
        
        cpu_cores = psutil.cpu_count()
        if max_workers > cpu_cores:
            max_workers = cpu_cores

        self.max_workers = max_workers
        self.loop = asyncio.get_event_loop()
        self.start_method = start_method

        self.context = multiprocessing.get_context(start_method)
        self.sem = BatchedSemaphore(max_workers)
        self.pool = ProcessPoolExecutor(max_workers=max_workers, mp_context=self.context)
        self.shutdown_task = None
        self.batch_by_stages = False

    async def execute_batches(self, batched_stages: List[Tuple[int, List[Any]]], execution_task: FunctionType) -> List[Tuple[str, Any]]:

        return await asyncio.gather(*[
            asyncio.create_task(
                self._execute_stage(
                    stage_name,
                    assigned_workers_count,
                    execution_task,
                    configs
                )
            ) for stage_name, assigned_workers_count, configs in batched_stages
        ])


    async def _execute_stage(
        self, 
        stage_name: str,
        assigned_workers_count: int,
        execution_task: FunctionType,
        configs: List[List[Any]]
    ) -> Tuple[str, Any]:

        await self.sem.acquire(assigned_workers_count)
        stage_results = await self.execute_stage_batch(
            execution_task,
            configs
        )

        self.sem.release(assigned_workers_count)

        return (stage_name, stage_results)

    async def execute_stage_batch(
        self, 
        execution_task: FunctionType,
        configs: List[Any]
    ):

        try:
            return await asyncio.gather(*[
                self.loop.run_in_executor(
                    self.pool,
                    execution_task,
                    config
                    
                ) for config in configs
            ])
        
        except BrokenProcessPool:
            raise ProcessKilledError()
        
        except KeyboardInterrupt:
            raise ProcessKilledError()

    def partion_stage_batches(self, stages: List[Any], ) -> List[Tuple[str, Any, int]]:

        # How many batches do we have? For example -> 5 stages over 4
        # CPUs means 2 batches. The first batch will assign one stage to
        # each core. The second will assign all four cores to the remaing
        # one stage.    

        batches = []

        if self.batch_by_stages is False:

            stages_count = len(stages)
            self.batch_sets_count = math.ceil(stages_count/self.max_workers)
            if self.batch_sets_count < 1:
                self.batch_sets_count = 1
            
            self.more_stages_than_cpus = stages_count/self.max_workers > 1

            if stages_count%self.max_workers > 0 and stages_count > self.max_workers:

                batch_size = self.max_workers
                workers_per_stage = int(self.max_workers/batch_size)
                batched_stages = [
                    [
                        workers_per_stage for _ in range(batch_size)
                    ] for _ in range(self.batch_sets_count - 1)
                ]

            else:
                batch_size = int(stages_count/self.batch_sets_count)
                workers_per_stage = int(self.max_workers/batch_size)
                batched_stages = [
                    [
                        workers_per_stage for _ in range(batch_size)
                    ] for _ in range(self.batch_sets_count)
                ]

            batches_count = len(batched_stages)

            # If we have a remainder batch - i.e. more stages than cores.
            last_batch = batched_stages[batches_count-1]

            last_batch_size = stages_count%self.max_workers
            if last_batch_size > 0 and self.more_stages_than_cpus:
                last_batch_workers = int(self.max_workers/last_batch_size)
                batched_stages.append([
                    last_batch_workers for _ in range(last_batch_size)
                ])

            last_batch = batched_stages[self.batch_sets_count-1]
            last_batch_size = len(last_batch)
            last_batch_remainder = self.max_workers%last_batch_size


            if last_batch_remainder > 0:
                for idx in range(last_batch_remainder):
                    last_batch[idx] += 1

            stage_idx = 0

            for batch in batched_stages:
                for stage_idx, stage_workers_count in enumerate(batch):

                    stage_name, stage = stages[stage_idx]

                    batches.append((
                        stage_name,
                        stage,
                        stage_workers_count
                    ))

        else:

            for stage_name, stage in stages:
                batches.append((
                    stage_name,
                    stage,
                    1
                ))

        return batches
    
    def partion_prioritized_stage_batches(self, transitions: List[Any], ) -> List[Tuple[str, str, str, int]]:

        # How many batches do we have? For example -> 5 stages over 4
        # CPUs means 2 batches. The first batch will assign one stage to
        # each core. The second will assign all four cores to the remaing
        # one stage.    

        batches = []
        seen_transitions: List[Any] = []


        sorted_transitions = list(sorted(
            transitions,
            key=lambda transition: transition.edge.source.priority_level.value
        ))


        bypass_partition_batch: List[Any] = []
        for transition in sorted_transitions:
            if transition.edge.skip_stage or transition.edge.source.allow_parallel is False:
                bypass_partition_batch.append((
                    transition.edge.source.name,
                    transition.edge.destination.name,
                    transition.edge.source.priority,
                    0
                ))

                seen_transitions.append(transition)

        batches.append(bypass_partition_batch)

        
        stages = {
            transition.edge.source.name: transition.edge.source for transition in sorted_transitions
        }

        parallel_stages_count = len([
            stage for stage in stages.values() if stage.allow_parallel
        ])

        stages_count = len(stages)

        auto_stages_count = len([
            stage for stage in stages.values() if stage.priority_level == StagePriority.AUTO
        ])

        if parallel_stages_count == 1:
            
            parallel_transitions = [
                transition for transition in sorted_transitions if transition.edge.skip_stage is False
            ]

            transition = parallel_transitions.pop()

            transition_group = [(
                transition.edge.source.name,
                transition.edge.destination.name,
                transition.edge.source.priority,
                transition.edge.source.workers
            )]

            return [transition_group]

        elif auto_stages_count == stages_count and parallel_stages_count > 0:
            transition_group = [
                (
                    transition.edge.source.name,
                    transition.edge.destination.name,
                    transition.edge.source.priority,
                    transition.edge.source.workers
                ) for transition in sorted_transitions if transition.edge.skip_stage is False
            ]

            return [transition_group]
            
        else:

            min_workers_counts: Dict[str, int] = {}
            max_workers_counts: Dict[str, int] = {}

            for transition in sorted_transitions:

                if transition.edge.skip_stage is False and transition.edge.source.allow_parallel:
                    worker_allocation_range: Tuple[int, int] = StagePriority.get_worker_allocation_range(
                        transition.edge.source.priority_level,
                        self.max_workers
                    )

                    minimum_workers, maximum_workers = worker_allocation_range
                    min_workers_counts[transition.edge.source.name] = minimum_workers
                    max_workers_counts[transition.edge.source.name] = maximum_workers

            for transition in sorted_transitions:
                
                if transition not in seen_transitions:


                    # So for example 8 - 4 = 4 we need another stage with 4
                    batch_workers_allocated = max_workers_counts.get(transition.edge.source.name)
                    transition_group = [(
                        transition.edge.source.name,
                        transition.edge.destination.name,
                        transition.edge.source.priority,
                        batch_workers_allocated
                    )]

                    for other_transition in sorted_transitions:
                        
                        if other_transition != transition and transition not in seen_transitions:
                            
                            transition_workers_allocated = max_workers_counts.get(other_transition.edge.source.name)
                            min_workers = min_workers_counts.get(other_transition.edge.source.name)

                            current_allocation = batch_workers_allocated + transition_workers_allocated

                            while current_allocation > self.max_workers and transition_workers_allocated >= min_workers:
                                transition_workers_allocated -= 1
                                current_allocation = batch_workers_allocated + transition_workers_allocated

                            if current_allocation <= self.max_workers and transition_workers_allocated > 0:

                                batch_workers_allocated += transition_workers_allocated
                                transition_group.append((
                                    other_transition.edge.source.name,
                                    other_transition.edge.destination.name,
                                    other_transition.edge.source.priority,
                                    transition_workers_allocated
                                ))

                                seen_transitions.append(other_transition)

                    batches.append(transition_group)
                    seen_transitions.append(transition)

        return batches

    async def shutdown(self):

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.pool.shutdown(cancel_futures=True)

            child_processes = active_children()
            for child in child_processes:
                child.kill()

    def close(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.pool.shutdown(cancel_futures=True)

            child_processes = active_children()
            for child in child_processes:
                child.kill()


from hedra.reporting.experiment.experiment_metrics_set import ExperimentMetricsSet
from hedra.reporting.metric.stage_metrics_summary import StageMetricsSummary
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from typing import Dict, Union


ExecutionResults = Dict[str, StageMetricsSummary]

GraphExecutionResults = Dict[str, Dict[str, Union[Dict[str, ExperimentMetricsSet], ExecutionResults, SystemMetricsSet]]]

SystemMetricsCollection = Dict[str, SystemMetricsSet]

GraphResults = Dict[str, Union[GraphExecutionResults, SystemMetricsCollection]]


from typing import (
    Any, 
    Dict, 
    List, 
    Callable, 
    Union
)
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.types.playwright import (
    MercuryPlaywrightClient,
    PlaywrightCommand,
    PlaywrightResult,
    Page,
    URL,
    Input,
    Options
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.client.store import ActionsStore
from hedra.logging import HedraLogger
from .base_client import BaseClient


class PlaywrightClient(BaseClient[MercuryPlaywrightClient, PlaywrightCommand, PlaywrightResult]):

    def __init__(self, config: Config) -> None:
        super().__init__()
        
        self.session = MercuryPlaywrightClient(
            concurrency=config.batch_size,
            group_size=config.group_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            )
        )

        self.request_type = RequestTypes.PLAYWRIGHT
        self.client_type = self.request_type.capitalize()

        self.actions: ActionsStore = None
        self.next_name = None
        self.intercept = False

        self.logger = HedraLogger()
        self.logger.initialize()

    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def goto(
        self, 
        url: str, 
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'goto',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def fill(
        self, 
        selector: str, 
        text: str,
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'fill',
            page=Page(selector=selector),
            input=Input(text=text),
            user=user,
            tags=tags,
            checks=checks     
        )

        return await self._execute_action(command)

    async def check(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'check',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def click(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'click',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def double_click(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'double_click',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def submit_event(
        self,
        selector: str,
        event: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'double_click',
            page=Page(selector=selector),
            options=Options(
                event=event, 
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def drag_and_drop(
        self,
        x_coordinate: int,
        y_coordinate: int,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'drag_and_drop',
            page=Page(
                x_coordinate=x_coordinate,
                y_coordinate=y_coordinate
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def switch_active_tab(
        self,
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'switch_active_tab',
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def evaluate_selector(
        self,
        selector: str,
        expression: str,
        args: List[Union[int, str, float, bool, None]]=[],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'evaluate_selector',
            page=Page(selector=selector),
            input=Input(
                expression=expression,
                args=args
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def evaluate_all_selectors(
        self,
        selector: str,
        expression: str,
        args: List[Union[int, str, float, bool, None]]=[],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'evaluate_all_selectors',
            page=Page(selector=selector),
            input=Input(
                expression=expression,
                args=args
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def evaluate_expression(
        self,
        expression: str,
        args: List[Union[int, str, float, bool, None]]=[],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'evaluate_expression',
            input=Input(
                expression=expression,
                args=args
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def evaluate_handle(
        self,
        expression: str,
        args: List[Union[int, str, float, bool, None]]=[],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'evaluate_handle',
            input=Input(
                expression=expression,
                args=args
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def exepect_console_message(
        self,
        expression: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'exepect_console_message',
            input=Input(expression=expression),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_download(
        self,
        expression: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_download',
            input=Input(expression=expression),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_event(
        self,
        event: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_event',
            options=Options(
                event=event, 
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_location(
        self,
        url: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_location',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_popup(
        self,
        expression: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_popup',
            input=Input(expression=expression),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_request(
        self,
        url: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_request',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_request_finished(
        self,
        url: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_request_finished',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def expect_response(
        self,
        url: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'expect_response',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def focus(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'focus',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def hover(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'hover',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_inner_html(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_inner_html',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_text(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_text',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_input_value(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_input_value',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def press_key(
        self,
        selector: str,
        key: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'press_key',
            page=Page(selector=selector),
            input=Input(key=key),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def verify_is_enabled(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'verify_is_enabled',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def verify_is_hidden(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'verify_is_hidden',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def verify_is_visible(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'verify_is_visible',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def verify_is_checked(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'verify_is_checked',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_content(
        self,
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_content',
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_element(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_element',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_all_elements(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_all_elements',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def reload_page(
        self,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'reload_page',
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def take_screenshot(
        self,
        screenshot_filepath: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'take_screenshot',
            input=Input(path=screenshot_filepath),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def select_option(
        self,
        selector: str,
        option: Union[str, int, bool, float, None],
        by_label: bool=False,
        by_value: bool=False,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'select_option',
            page=Page(selector=selector),
            input=Input(
                option=option,
                by_label=by_label,
                by_value=by_value
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def set_checked(
        self,
        selector: str,
        checked: bool,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'set_checked',
            page=Page(selector=selector),
            options=Options(
                is_checked=checked,
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def set_default_timeout(
        self,
        timeout: float,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'set_default_timeout',
            options=Options(
                timeout=timeout,
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def set_navigation_timeout(
        self,
        timeout: float,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'set_navigation_timeout',
            options=Options(
                timeout=timeout,
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def set_http_headers(
        self,
        headers: Dict[str, str],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'set_http_headers',
            url=URL(headers=headers),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def tap(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'tap',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_text_content(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_text_content',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_page_title(
        self,
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_page_title',
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def input_text(
        self,
        selector: str,
        text: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'input_text',
            page=Page(selector=selector),
            input=Input(text=text),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def uncheck(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'uncheck',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def wait_for_event(
        self,
        event: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_event',
            options=Options(
                event=event,
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def wait_for_function(
        self,
        expression: str,
        args: List[Union[int, str, float, bool, None]]=[],
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_function',
            input=Input(
                expression=expression,
                args=args
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def wait_for_page_load_state(
        self,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_page_load_state',
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def wait_for_selector(
        self,
        selector: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_selector',
            page=Page(selector=selector),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def wait_for_timeout(
        self,
        timeout: float,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_timeout',
            options=Options(
                timeout=timeout,
                **extra
            ),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)
    
    async def wait_for_url(
        self,
        url: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'wait_for_url',
            url=URL(location=url),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def switch_frame(
        self,
        url: str=None,
        frame: int=None,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'switch_frame',
            url=URL(location=url),
            page=Page(frame=frame),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_frames(
        self,
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_frames',
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def get_attribute(
        self,
        selector: str,
        attribute: str,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'get_attribute',
            page=Page(
                selector=selector,
                attribute=attribute
            ),
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def go_back_page(
        self,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'go_back_page',
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)

    async def go_forward_page(
        self,
        extra: Dict[str, Any]={}, 
        user: str=None,
        tags: List[Dict[str, str]]=[],
        checks: List[Callable]=[]
    ):
        command = PlaywrightCommand(
            self.next_name,
            'go_forward_page',
            options=Options(**extra),
            user=user,
            tags=tags,
            checks=checks
        )

        return await self._execute_action(command)



from typing import Callable, Coroutine, Any


class BaseWrapper:
    
    def __init__(self) -> None:
        self.setup: Callable[
            [],
            Coroutine[
                Any,
                Any,
                None
            ]
        ] = None


from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.hooks import (
    action
)


class ExecuteGraphQLHttp2Stage(Execute):

    @action()
    async def http2_query(self):
        return await self.client.graphqlh2.query(
            """
            query <query_name> {
                ...
            }
            """
        )


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    AnyHttpUrl,
    Json
)


from typing import List, Dict, Optional, Union


class HTTP3ActionTag(BaseModel):
    name: StrictStr
    value: StrictStr


class HTTP3ActionValidator(BaseModel):
    engine: StrictStr
    name: StrictStr
    url: AnyHttpUrl
    method: StrictStr='GET'
    headers: Dict[StrictStr, StrictStr]={}
    params: Optional[Dict[StrictStr, Union[StrictInt, StrictStr, StrictFloat]]]
    data: Optional[Union[StrictStr, Json]]
    weight: Optional[Union[StrictInt, StrictFloat]]
    order: Optional[StrictInt]
    user: Optional[StrictStr]
    tags: List[HTTP3ActionTag]=[]


from __future__  import annotations
import inspect
import asyncio
import traceback
from collections import defaultdict
from typing import List, Dict, Any, Union
from hedra.core.graphs.transitions.common.base_edge import BaseEdge
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.submit.submit import Submit
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.registrar import registrar
from hedra.reporting.experiment.experiment_metrics_set import ExperimentMetricsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.metric.custom_metric import CustomMetric
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.reporting.metric.stage_metrics_summary import StageMetricsSummary
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from hedra.reporting.system.system_metrics_set_types import MonitorGroup


CustomMetricSet = Dict[str, Dict[str, CustomMetric]]
MetricsSetGroup = Dict[str, Union[str, Dict[str, StageMetricsSummary]]]


class SubmitEdge(BaseEdge[Submit]):

    def __init__(self, source: Submit, destination: BaseEdge[Stage]) -> None:
        super(
            SubmitEdge,
            self
        ).__init__(
            source,
            destination
        )
        
        self.requires = [
            'analyze_stage_session_total',
            'analyze_stage_events',
            'analyze_stage_summary_metrics',
            'session_stage_monitors',
            'submit_stage_session_total'
        ]
        self.provides = [
            'submit_stage_metrics',
            'submit_stage_streamed_metrics',
            'submit_stage_experiment_metrics',
            'submit_stage_monitors',
            'session_stage_monitors',
            'submit_stage_system_metrics'
        ]

    async def transition(self):
        
        try:
            self.source.state = StageStates.SUBMITTING
            
            self.source.context.update(self.edge_data)

            for event in self.source.dispatcher.events_by_name.values():
                event.source.stage_instance = self.source
                event.context.update(self.edge_data)
                
                if event.source.context:
                    event.source.context.update(self.edge_data)

            if self.timeout and self.skip_stage is False:
                await asyncio.wait_for(self.source.run(), timeout=self.timeout)
            
            elif self.skip_stage is False:
                await self.source.run()
            
            for provided in self.provides:
                self.edge_data[provided] = self.source.context[provided]

            self._update(self.destination)

            self.source.state = StageStates.SUBMITTED
            self.destination.state = StageStates.SUBMITTED

            if self.destination.context is None:
                self.destination.context = SimpleContext()

            self.visited.append(self.source.name)

        except Exception as edge_exception:
            self.exception = edge_exception

        return None, self.destination.stage_type

    def _update(self, destination: Stage):

        for edge_name in self.history:

            history = self.history[edge_name]

            if self.next_history.get(edge_name) is None:
                self.next_history[edge_name] = {}

            self.next_history[edge_name].update(history)

        session_stage_monitors: MonitorGroup = self.edge_data['session_stage_monitors']
        session_stage_monitors.update(
            self.edge_data['submit_stage_monitors']
        )

        if self.skip_stage is False:
            self.next_history.update({
                (self.source.name, destination.name): {
                    'submit_stage_metrics': self.edge_data['submit_stage_metrics'],
                    'submit_stage_experiment_metrics': self.edge_data['submit_stage_experiment_metrics'],
                    'submit_stage_streamed_metrics': self.edge_data['submit_stage_streamed_metrics'],
                    'submit_stage_system_metrics': self.edge_data['submit_stage_system_metrics']
                }
            })

    def split(self, edges: List[SubmitEdge]) -> None:
        
        submit_stage_config: Dict[str, Any] = self.source.to_copy_dict()
        submit_stage_copy: Submit = type(self.source.name, (Submit, ), self.source.__dict__)()
        
        for copied_attribute_name, copied_attribute_value in submit_stage_config.items():
            if inspect.ismethod(copied_attribute_value) is False:
                setattr(
                    submit_stage_copy, 
                    copied_attribute_name, 
                    copied_attribute_value
                )

        user_hooks: Dict[str, Dict[str, Hook]] = defaultdict(dict)
        for hooks in registrar.all.values():
            for hook in hooks:
                if hasattr(self.source, hook.shortname) and not hasattr(Submit, hook.shortname):
                    user_hooks[self.source.name][hook.shortname] = hook._call
        
        submit_stage_copy.dispatcher = self.source.dispatcher.copy()

        for event in submit_stage_copy.dispatcher.events_by_name.values():
            event.source.stage_instance = submit_stage_copy

        submit_stage_copy.context = SimpleContext()
        for event in submit_stage_copy.dispatcher.events_by_name.values():
            event.source.stage_instance = submit_stage_copy 
            event.source.stage_instance.context = submit_stage_copy.context
            event.source.context = submit_stage_copy.context

            if event.source.shortname in user_hooks[submit_stage_copy.name]:
                hook_call = user_hooks[submit_stage_copy.name].get(event.source.shortname)

                hook_call = hook_call.__get__(submit_stage_copy, submit_stage_copy.__class__)
                setattr(submit_stage_copy, event.source.shortname, hook_call)

                event.source._call = hook_call

            else:            
                event.source._call = getattr(submit_stage_copy, event.source.shortname)
                event.source._call = event.source._call.__get__(submit_stage_copy, submit_stage_copy.__class__)
                setattr(submit_stage_copy, event.source.shortname, event.source._call)
 
        self.source = submit_stage_copy

        transition_idxs = [edge.transition_idx for edge in edges]
        
        if self.transition_idx != min(transition_idxs):
            self.skip_stage = True

    def setup(self):
        
        events: List[BaseProcessedResult] = []
        metrics: List[MetricsSet] = []
        streamed_metrics: Dict[str, StageStreamsSet] = {}
        system_metrics: List[SystemMetricsSet] = []
        experiments: List[ExperimentMetricsSet] = []
        session_total: int = 0
        session_stage_monitors: MonitorGroup = {}

        for source_stage, destination_stage in self.history:
            if destination_stage == self.source.name:

                previous_history: Dict[str, Any] = self.history[(source_stage, self.source.name)]
            
                analyze_stage_summary_metrics: MetricsSetGroup = previous_history.get(
                    'analyze_stage_summary_metrics'
                )
                analyze_stage_events: List[BaseProcessedResult] = previous_history.get(
                    'analyze_stage_events',
                    []
                )

                events.extend(analyze_stage_events)


                stage_metrics_summaries = analyze_stage_summary_metrics.get('stages', {})
                for stage_metrics_summary in stage_metrics_summaries.values():
                    metrics.extend(list(
                        stage_metrics_summary.metrics_sets.values()
                    ))

                    session_total += stage_metrics_summary.stage_metrics.total

                experiment_metrics_sets = analyze_stage_summary_metrics.get('experiment_metrics_sets', {})
                experiments.extend(list(
                    experiment_metrics_sets.values()
                ))

                for stage_metrics_summary in stage_metrics_summaries.values():
                    streams = stage_metrics_summary.stage_streamed_analytics

                    if streams and len(streams) > 0:
                        streamed_metrics[stage_metrics_summary.stage_metrics.name] = stage_metrics_summary.streams

                analyze_stage_system_metrics = analyze_stage_summary_metrics.get('system_metrics')
                if analyze_stage_system_metrics:
                    system_metrics.append(analyze_stage_system_metrics)

                stage_monitors = previous_history.get('session_stage_monitors')
                if stage_monitors:
                    session_stage_monitors.update(stage_monitors)

        self.edge_data = {
            'session_stage_monitors': session_stage_monitors,
            'submit_stage_events': events,
            'submit_stage_experiment_metrics': experiments,
            'submit_stage_streamed_metrics': streamed_metrics,
            'submit_stage_system_metrics': system_metrics,
            'submit_stage_summary_metrics': metrics,
            'submit_stage_session_total': session_total
        }



import asyncio
from hedra.distributed.models.http import (
    Limit,
    HTTPMessage,
    Request
)
from types import TracebackType
from typing import (
    Optional, 
    Type
)
from .base_limiter import BaseLimiter


class TokenBucketLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check"
    )

    def __init__(
        self, 
        limit: Limit
    ) -> None:
        super().__init__(
            limit.max_requests,
            limit.period,
            reject_requests=limit.reject_requests
        )

        self._level = limit.max_requests
        self._last_check = self._loop.time()

    def has_capacity(self, amount: float = 1) -> bool:

        if self._level < self.max_rate:
            current_time = self._loop.time()
            delta = self._rate_per_sec * (current_time - self._last_check)
            self._level = min(self.max_rate, self._level + delta)
            self._last_check = current_time

        requested_amount = self._level - amount
        if requested_amount > 0 or self._level >= self.max_rate:
            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break


        return amount < self._level
    
    async def acquire(
        self, 
        amount: float = 1
    ):
        
        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")


        task = asyncio.current_task(
            loop=self._loop
        )

        assert task is not None

        rejected = False

        if not self.has_capacity(amount) and self._reject_requests:
            return True

        while not self.has_capacity(amount):

            fut = self._loop.create_future()

            try:

                self._waiters[task] = fut
                await asyncio.wait_for(
                    asyncio.shield(fut), 
                    timeout=(1 / self._rate_per_sec * amount)
                )

            except asyncio.TimeoutError:
                pass

            fut.cancel()
            if self._reject_requests:
                rejected = True

        self._waiters.pop(task, None)
        self._level -= amount

        return rejected

    async def reject(
        self,
        request: Request,
        transport: asyncio.Transport
    ):
        if transport.is_closing() is False:

            server_error_respnse = HTTPMessage(
                path=request.path,
                status=429,
                error='Too Many Requests',
                method=request.method
            )

            transport.write(server_error_respnse.prepare_response())


    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        return None


import click
from hedra.cli.plugin import create_plugin


@click.group(help='Commands for creating and managing Hedra plugins.')
def plugin():
    pass



@plugin.command()
@click.argument('plugin_type')
@click.argument('path')
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
def create(plugin_type: str, path: str, log_level: str):
    create_plugin(plugin_type, path, log_level)


from __future__ import annotations
import socket
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType


class ARecordData(RecordData):
    '''A record'''

    def __init__(self, data: str):
        super().__init__(
            RecordType.A,
            data=data
        )

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, ARecordData]:
        
        ip = socket.inet_ntoa(
            data[cursor_position:cursor_position + size]
        )
        
        return cursor_position + size, ARecordData(ip)

    def dump(
        self, 
        names: Dict[str, int], 
        offset: int
    ) -> Iterable[bytes]:
        yield socket.inet_aton(self.data)



import datetime
from typing import List
from hedra.logging import HedraLogger
from .action import RepoAction
from .config import RepoConfig


class Fetch(RepoAction):

    def __init__(self, config: RepoConfig, graph_files: List[str]) -> None:
        super(
            Fetch,
            self
        ).__init__(config, graph_files)

    def execute(self):
        self.logger.hedra.sync.debug(f'FetchAction: {self.action_id} - Cloning from remote - {self.config.uri} - to path - {self.config.path}')
        self._pull_from_remote()

        self.logger.hedra.sync.debug(f'FetchAction: {self.action_id} - Checking out branch - {self.config.branch} - and updating remote')
        self._checkout()

        self.logger.hedra.sync.debug(f'FetchAction: {self.action_id} - Updating .gitignore')
        self._update_ignore()



import re

class MemoryParser:

    def __init__(self, time_amount: str) -> None:
        self.UNITS = {
            'kb':'kilobytes', 
            'mb':'megabytes', 
            'gb':'gigabytes'
        }

        self._conversion_table = {
            'kilobytes': {
                'kilobytes': 1,
                'megabytes': 1/1024,
                'gigabytes': 1/(1024**2)
            },
            'megabytes': {
                'kilobytes': 1024,
                'megabytes': 1,
                'gigabytes': 1/1024
            },
            'gigabytes': {
                'kilobytes': 1024**2,
                'megabytes': 1024,
                'gigabytes': 1
            }
        }
        
        
        parsed_size = {
            self.UNITS.get(
                m.group(
                    'unit'
                ).lower(), 
                'megabytes'
            ): float(m.group('val'))
            for m in re.finditer(
                r'(?P<val>\d+(\.\d+)?)(?P<unit>[smhdw]?)', 
                time_amount, 
                flags=re.I
            )
        }

        self.unit = list(parsed_size.keys()).pop()
        self.size = parsed_size.pop(self.unit)

    def kilobytes(self, accuracy: int = 2):
        conversion_amount = self._conversion_table.get(
            self.unit,
            {}
        ).get(
            'kilobytes',
            1
        )

        return round(
            self.size * conversion_amount,
            accuracy
        )

    def megabytes(self, accuracy: int = 2):
        conversion_amount = self._conversion_table.get(
            self.unit,
            {}
        ).get(
            'megabytes',
            1
        )

        return round(
            self.size * conversion_amount,
            accuracy
        )
    
    def gigabytes(self, accuracy: int = 2):
        conversion_amount = self._conversion_table.get(
            self.unit,
            {}
        ).get(
            'gigabytes',
            1
        )
        

        return round(
            self.size * conversion_amount,
            accuracy
        )


import asyncio
import psutil
import traceback
from collections import defaultdict
from typing_extensions import TypeVarTuple, Unpack
from typing import (
    Dict, 
    Generic, 
    List, 
    Any, 
    Optional, 
    Union
)
from hedra.core.experiments.experiment import Experiment
from hedra.core.engines.client.client import Client
from hedra.core.engines.client.config import Config
from hedra.core.engines.client.tracing_config import TracingConfig
from hedra.core.hooks.types.condition.decorator import condition
from hedra.core.hooks.types.context.decorator import context
from hedra.core.hooks.types.event.decorator import event
from hedra.core.hooks.types.transform.decorator import transform
from hedra.core.engines.types.common.action_registry import actions_registry
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.logging import HedraLogger
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.core.personas.types import PersonaTypesMap
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.plugins.types.plugin_types import PluginType
from hedra.core.graphs.stages.base.stage import Stage
from .exceptions import HookSetupError

try:

    from playwright.async_api import Geolocation

except Exception:
    Geolocation = None


T = TypeVarTuple('T')


class SetupCall:

    def __init__(self, hook: Hook, config: Config, retries: int=1) -> None:
        self.hook = hook
        self.config = config
        self.hook_name = self.hook.hook_type.name.capitalize()
        self.exception = None
        self.action_store = None
        self.retries = retries
        self.current_try = 1
        self.logger = HedraLogger()
        self.logger.initialize()
        self.metadata_string:str = None
        self.error_traceback: str = None

    async def setup(self):

        for _ in range(self.retries):
            try:

                self.hook.stage_instance.client._config = self.config
                self.hook.stage_instance.client.set_mutations()
                await self.hook.call()

            except Exception as setup_exception:
                self.exception = setup_exception

                self.error_traceback = str(traceback.format_exc())
                await self.logger.filesystem.aio['hedra.core'].error(f'{self.metadata_string} - Encountered connection validation error - {str(setup_exception)} - {self.hook_name} Hook: {self.hook.name}:{self.hook.hook_id}')

                if self.current_try >= self.retries:
                    self.action_store.waiter.set_result(None)
                    break

                else:
                    self.current_try += 1



class Setup(Stage, Generic[Unpack[T]]):
    experiment: Optional[Experiment] = None
    stage_type=StageTypes.SETUP
    log_level='info'
    persona_type='default'
    total_time='1m'
    batch_size=1000
    batch_interval=0
    action_interval=0
    batch_gradient=0.1
    cpus=int(psutil.cpu_count(logical=False))
    no_run_visuals=False
    graceful_stop=1
    connect_timeout=10
    request_timeout=60
    reset_connections=False
    apply_to_stages=[]
    browser_type: str='chromium'
    device_type: str=None
    locale: str=None
    geolocation: Geolocation=None
    permissions: List[str]=[]
    playwright_options: Dict[str, Any]={}
    tracing: TracingConfig=None
    priority: Optional[str]=None
    actions_filepaths: Optional[Dict[str, str]]=None
    retries: int=0

    
    def __init__(self) -> None:
        super().__init__()
        self.generation_setup_candidates = 0
        self.stages: Dict[str, Stage] = {}
        self.accepted_hook_types = [ 
            HookType.CONDITION,
            HookType.CONTEXT,
            HookType.EVENT, 
            HookType.TRANSFORM 
        ]

        self.persona_types = PersonaTypesMap()
        self.config = Config(
            log_level=self.log_level,
            persona_type=self.persona_types[self.persona_type],
            total_time=self.total_time,
            batch_size=self.batch_size,
            batch_interval=self.batch_interval,
            action_interval=self.action_interval,
            batch_gradient=self.batch_gradient,
            cpus=self.cpus,
            no_run_visuals=self.no_run_visuals,
            connect_timeout=self.connect_timeout,
            request_timeout=self.request_timeout,
            graceful_stop=self.graceful_stop,
            reset_connections=self.reset_connections,
            browser_type=self.browser_type,
            device_type=self.device_type,
            locale=self.locale,
            geolocation=self.geolocation,
            permissions=self.permissions,
            playwright_options=self.playwright_options,
            tracing=self.tracing,
            actions_filepaths=self.actions_filepaths
        )

        self.client = Client(
            self.graph_name,
            self.graph_id,
            self.name,
            self.stage_id
        )
        self.client._config = self.config
    
        self.experiment = self.experiment
        if self.experiment:
            self.experiment.source_batch_size = self.batch_size

        self.source_internal_events = [
            'collect_target_stages'
        ]

        self.internal_events = [
            'collect_target_stages',
            'configure_target_stages',
            'collect_action_hooks',
            'check_actions_setup_needed',
            'setup_action',
            'collect_task_hooks',
            'check_tasks_setup_needed',
            'setup_task',
            'apply_channels',
            'complete'
        ]

        self.tracing = self.tracing
        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.stage_retries = self.retries

    @Internal()
    async def run(self):
        await self.setup_events()
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)

    @Internal()
    async def run_internal(self):
        await self.setup_events()
        
        initial_events = dict(**self.dispatcher.initial_events)
        self.dispatcher.skip_list.extend([
            stage_event.event_name for stage_event in self.dispatcher.events_by_name.values() if stage_event.source.shortname not in self.internal_events
        ])

        self.dispatcher.initial_events = initial_events
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)
    
    @context()
    async def collect_target_stages(
        self,
        setup_stage_target_stages: Dict[str, Stage]={},
        setup_stage_target_config: Config=None,
        setup_stage_is_primary_thread: bool = True
    ):
  
        bypass_connection_validation = self.core_config.get('bypass_connection_validation', False)
        connection_validation_retries = self.core_config.get('connection_validation_retries', 3)

        self.context.ignore_serialization_filters = [
            'session_stage_monitors'
        ]

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting setup')

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        main_monitor_name = f'{self.name}.main'

        if setup_stage_is_primary_thread:
            await cpu_monitor.start_background_monitor(main_monitor_name)
            await memory_monitor.start_background_monitor(main_monitor_name)

        
        return {
            'setup_stage_experiment_config': {},
            'setup_stage_target_stages_count': len(setup_stage_target_stages),
            'setup_stage_target_config': setup_stage_target_config,
            'setup_stage_target_stages': setup_stage_target_stages,
            'setup_stage_monitors': {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            },
            'bypass_connection_validation': bypass_connection_validation,
            'connection_validation_retries': connection_validation_retries,
            'setup_stage_is_primary_thread': setup_stage_is_primary_thread
        }

    @event('collect_target_stages')
    async def configure_target_stages(
        self, 
        setup_stage_target_stages: Dict[str, Stage]={},
        setup_stage_target_config: Config=None,
        setup_stage_is_primary_thread: bool=True,
        setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]]={},
    ):
        execute_stage_names = ', '.join(list(setup_stage_target_stages.keys()))

        setup_stage_configs = {}

        await self.logger.spinner.append_message(f'Setting up - {execute_stage_names}')

        execute_stage_id = 1

        if setup_stage_is_primary_thread and self.experiment:
            self.experiment.assign_weights()
        
        for execute_stage_name, execute_stage in setup_stage_target_stages.items():

            execute_stage.execution_stage_id = execute_stage_id
            execute_stage.execute_setup_stage = self.name

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Execute stage - {execute_stage_name} - assigned stage order id - {execute_stage_id}')

            execute_stage_id += 1

            persona_plugins: Dict[str, PersonaPlugin] = self.plugins_by_type[PluginType.PERSONA]
            for plugin_name, plugin in persona_plugins.items():
                plugin.name = plugin_name
                self.persona_types.types[plugin_name] = plugin
                execute_stage.plugins[plugin_name] = plugin
                
                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Persona plugin - {plugin.name} - for Execute stae - {execute_stage_name}')
   
            engine_plugins: Dict[str, EnginePlugin] = self.plugins_by_type[PluginType.ENGINE]

            for plugin_name, plugin in engine_plugins.items():
                execute_stage.client.plugin[plugin_name] = plugin(setup_stage_target_config)
                plugin.name = plugin_name
                execute_stage.plugins[plugin_name] = plugin
                self.plugins_by_type[plugin.type][plugin_name] = plugin

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Engine plugin - {plugin.name} - for Execute stage - {execute_stage_name}')

            existing_stage_config = setup_stage_configs.get(execute_stage_name)
            if existing_stage_config:
                return {
                    'setup_stage_configs': setup_stage_configs,
                    'setup_stage_target_stages': setup_stage_target_stages,
                    'setup_stage_experiment_config': setup_stage_experiment_config
                }

            config_copy = setup_stage_target_config.copy()

            if self.tracing:
                config_copy.tracing = self.tracing

            if self.experiment and self.experiment.is_variant(execute_stage_name):


                variant = self.experiment.get_variant(execute_stage_name)

                experiment = {
                    'experiment_name': self.experiment.experiment_name,
                    'random': self.experiment.random,
                    'weight': variant.weight
                }

                if setup_stage_is_primary_thread:
                    config_copy.batch_size = self.experiment.get_variant_batch_size(
                        execute_stage_name
                    )

                    distribution = self.experiment.calculate_distribution(
                        execute_stage_name,
                        config_copy.batch_size
                    )

                    if distribution is not None:


                        experiment.update({
                            'distribution_type': variant.distribution.selected_distribution,
                            'distribution': distribution,
                            'intervals': variant.intervals,
                            'interval_duration': round(
                                config_copy.total_time/(variant.intervals - 1), 
                                2
                            )
                        })

                        config_copy.experiment = experiment
                        config_copy.persona_type = self.persona_types['approx-dist']

                if variant.mutations:
                    config_copy.mutations = variant.get_mutations()

                setup_stage_experiment_config[execute_stage_name] = experiment

            setup_stage_configs[execute_stage_name] = config_copy
            setup_stage_target_stages[execute_stage_name] = execute_stage

        return  {
            'setup_stage_configs': setup_stage_configs,
            'setup_stage_target_stages': setup_stage_target_stages,
            'setup_stage_experiment_config': setup_stage_experiment_config
        }

    @event('configure_target_stages')
    async def collect_action_hooks(
        self,
        setup_stage_target_stages: Dict[str, Stage]={}
    ):
        actions: List[ActionHook] = []
        for execute_stage in setup_stage_target_stages.values():
            actions.extend(execute_stage.hooks[HookType.ACTION])

        return {
            'setup_stage_actions': actions
        }

    
    @condition('collect_action_hooks')
    async def check_actions_setup_needed(
        self, 
        setup_stage_actions: List[ActionHook]=[]
    ):
        
        return {
            'setup_stage_has_actions': len(setup_stage_actions) > 0
        }
        
    @transform('check_actions_setup_needed')
    async def setup_action(
        self,
        setup_stage_configs: Dict[str, Config] = {},
        setup_stage_actions: ActionHook=None,
        setup_stage_has_actions: bool = False,
        bypass_connection_validation: bool=False,
        connection_validation_retries: int=3,
        setup_stage_target_config: Config=None
    ):
            if setup_stage_has_actions and isinstance(setup_stage_actions, ActionHook) and setup_stage_actions.skip is False:

                hook = setup_stage_actions
                hook.skip = setup_stage_actions.skip
                hook.stage_instance.client.next_name = hook.name
                hook.stage_instance.client.intercept = True

                config_copy = setup_stage_configs.get(
                    hook.stage
                )
                
                hook.stage_instance.client._config = config_copy

                execute_stage_name = hook.stage_instance.name

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Client intercept set to {hook.stage_instance.client.intercept} - Action calls for client id - {hook.stage_instance.client.client_id} - will be suspended on execution')

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Setting up Action - {hook.name}:{hook.hook_id} - for Execute stage - {execute_stage_name}')
                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Preparing Action hook - {hook.name}:{hook.hook_id} - for suspension - Execute stage - {execute_stage_name}')

                hook.stage_instance.client.actions.set_waiter(hook.stage_instance.name)
                
                setup_call = SetupCall(
                    hook, 
                    config_copy, 
                    retries=connection_validation_retries
                )

                setup_call.metadata_string = self.metadata_string
                setup_call.action_store = hook.stage_instance.client.actions

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Executing Action hook call - {hook.name}:{hook.hook_id} - Execute stage - {execute_stage_name}')

                task = asyncio.create_task(setup_call.setup())
                await hook.stage_instance.client.actions.wait_for_ready(setup_call)   

                await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Exiting suspension for Action - {hook.name}:{hook.hook_id} - Execute stage - {execute_stage_name}')

                action = None
                session = None

                try:
                    if setup_call.exception:
                        raise HookSetupError(
                            hook, 
                            HookType.ACTION, 
                            str(setup_call.exception)
                        )

                    task.cancel()
                    if task.cancelled() is False:
                        await asyncio.wait_for(task, timeout=0.1)

                except HookSetupError as hook_setup_exception:
                    if bypass_connection_validation:

                        action.hook_type = HookType.TASK

                        hook.stage_instance.hooks[HookType.TASK].append(hook)
                        action_idx = hook.stage_instance.hooks[HookType.ACTION].index(hook)
                        hook.stage_instance.hooks[HookType.ACTION].pop(action_idx)

                    else:
                        raise hook_setup_exception

                except asyncio.InvalidStateError:
                    pass

                except asyncio.CancelledError:
                    pass

                except asyncio.TimeoutError:
                    pass
                
                action, session = hook.stage_instance.client.actions.get(
                    hook.stage_instance.name,
                    hook.name
                )

                await session.set_pool(setup_stage_target_config.batch_size)

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Successfully retrieved prepared Action and Session for action - {hook.name}:{action.action_id} - Execute stage - {execute_stage_name}')
                    
                if len(hook.before) > 0:
                    action.hooks.before = hook.before

                if len(hook.after) > 0:
                    action.hooks.after = hook.after

                if len(hook.checks) > 0:
                    action.hooks.checks = hook.checks
                    
                hook.session = session
                hook.action = action  

                actions_registry[hook.name] = action

                return {
                    'setup_stage_actions': setup_stage_actions
                }

    @event('setup_action')
    async def collect_task_hooks(
        self,
        setup_stage_target_stages: Dict[str, Stage]={}
    ):
        tasks: List[TaskHook] = []
        for execute_stage in setup_stage_target_stages.values():
            tasks.extend(execute_stage.hooks[HookType.TASK])

        return {
            'setup_stage_tasks': tasks
        }
    
    @condition('collect_task_hooks')
    async def check_tasks_setup_needed(
        self, 
        setup_stage_tasks: List[ActionHook]=[]
    ):
        return {
            'setup_stage_has_tasks': len(setup_stage_tasks) > 0
        }

    @transform('check_tasks_setup_needed')
    async def setup_task(
        self,
        setup_stage_tasks: TaskHook=None,
        setup_stage_has_tasks: bool=False,
        setup_stage_configs: Dict[str, Config] = {},
    ):
        if setup_stage_has_tasks and isinstance(setup_stage_tasks, TaskHook) and setup_stage_tasks.skip is False:
            hook = setup_stage_tasks
            execute_stage: Stage = hook.stage_instance
            execute_stage.client.next_name = hook.name
            execute_stage.client.intercept = True

            config_copy = setup_stage_configs.get(
                hook.stage
            )
                
            execute_stage.client._config = config_copy
            execute_stage.client.set_mutations()

            execute_stage_name = hook.stage_instance.name

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Loading Task hook - {hook.name}:{hook.hook_id} - to Execute stage - {execute_stage_name}')

            execute_stage.client.next_name = hook.name
            task, session = execute_stage.client.task.call(
                hook.call,
                env=hook.metadata.env,
                user=hook.metadata.user,
                tags=hook.metadata.tags
            )
            
            await session.set_pool(config_copy.batch_size)

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Successfully retrieved task and session for Task - {hook.name}:{task.action_id} - Execute stage - {execute_stage_name}')

            if len(hook.before) > 0:
                task.hooks.before = hook.before

            if len(hook.after) > 0:
                task.hooks.after = hook.after

            if len(hook.checks) > 0:
                task.hooks.checks = hook.checks

            
            if hook.is_notifier:
                task.hooks.notify = hook.is_notifier
                task.hooks.channels = hook.channels
                task.hooks.listeners = {
                    listener.shortname: listener for listener in hook.listeners
                }

            elif hook.is_listener:
                task.hooks.listen = hook.is_listener

            hook.session = session
            hook.action = task  

            actions_registry[hook.name] = task

            return {
                'setup_stage_tasks': setup_stage_tasks
            }
        

    @context('setup_task')
    async def apply_channels(
        self,
        setup_stage_configs: Dict[str, Config] = {},
        setup_stage_target_stages: Dict[str, Stage]=[],
        setup_stage_actions: List[ActionHook]=[],
        setup_stage_tasks: List[TaskHook]=[],
        setup_stage_target_config: Config=None,
        setup_stage_experiment_config: Dict[str, Union[str, int, List[float]]]={},
        setup_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={},
        setup_stage_is_primary_thread: bool = True
    ):
        
        stage_cpu_monitor: CPUMonitor = setup_stage_monitors.get('cpu')
        stage_memory_monitor: MemoryMonitor = setup_stage_monitors.get('memory')

        if setup_stage_is_primary_thread:
            main_monitor_name = f'{self.name}.main'
    
            await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
            await stage_memory_monitor.stop_background_monitor(main_monitor_name)

            stage_cpu_monitor.close()
            stage_memory_monitor.close()

            stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
            stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]      

        actions_by_stage = defaultdict(list)
        tasks_by_stage = defaultdict(list)

        for action in setup_stage_actions:
            actions_by_stage[action.stage].append(action)

        for task in setup_stage_tasks:
            tasks_by_stage[task.stage].append(task)

        execute_stage_setup_hooks = defaultdict(list)

        for execute_stage in setup_stage_target_stages.values():

            execute_stage.client.intercept = False
            execute_stage.hooks[HookType.ACTION] = actions_by_stage[execute_stage.name]
            execute_stage.hooks[HookType.TASK] = tasks_by_stage[execute_stage.name]
            execute_stage.context['execute_stage_setup_hooks'] = [
                *actions_by_stage[execute_stage.name],
                *tasks_by_stage[execute_stage.name]
            ]

            execute_stage_setup_hooks[execute_stage.name].extend([
                *actions_by_stage[execute_stage.name],
                *tasks_by_stage[execute_stage.name]
            ])

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Client intercept set to {execute_stage.client.intercept} - Action calls for client id - {execute_stage.client.client_id} - will not be suspended on execution')

            self.stages[execute_stage.name] = execute_stage

            actions_generated_count = len(execute_stage.hooks[HookType.ACTION])
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Generated - {actions_generated_count} - Actions for Execute stage - {execute_stage.name}')

            tasks_generated_count = len(execute_stage.hooks[HookType.TASK])
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Generated - {tasks_generated_count} - Tasks for Execute stage - {execute_stage.name}')

        return {
            'setup_stage_experiment_config': setup_stage_experiment_config,
            'setup_stage_configs': setup_stage_configs,
            'execute_stage_setup_by': self.name,
            'execute_stage_setup_hooks': execute_stage_setup_hooks,
            'execute_stage_setup_config': setup_stage_target_config,
            'setup_stage_ready_stages': setup_stage_target_stages,
            'setup_stage_target_stages': setup_stage_target_stages,
            'setup_stage_monitors': {
                self.name: {
                    'cpu': stage_cpu_monitor,
                    'memory': stage_memory_monitor
                }
            },
        }

    @event('apply_channels')
    async def complete(self):
        return {}




import inspect
from typing import Coroutine, Dict, List, Any
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.types import RequestTypes


class Task(BaseAction):

    __slots__ = (
        'action_id',
        'protocols', 
        'name', 
        'is_setup', 
        'metadata', 
        'hooks',
        'type',
        'source',
        'execute',
        'event',
        'args',
        'params',
        'task_args',
        'mutations'
    )

    def __init__(
        self, 
        name: str,
        task_action: Coroutine,
        source: str=None,
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ):
        super(Task, self).__init__(
            name,
            user,
            tags
        )

        self.type = RequestTypes.TASK
        self.source = source
        self.execute = task_action
        self.hooks: Hooks[Task] = Hooks()


        self.args = inspect.signature(task_action)
        self.params = self.args.parameters
        self.task_args: Dict[str, Any] = {}



import asyncio
import functools
import json
import signal
import uuid
import psutil
from typing import List, Dict, Union
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from hedra.reporting.types import ReporterTypes
from .aws_lambda_config import AWSLambdaConfig

try:
    import boto3
    has_connector=True
except Exception:
    boto3 = None
    has_connector=False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class AWSLambda:

    def __init__(self, config: AWSLambdaConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name

        self.events_lambda_name = config.events_lambda
        self.streams_lambda_name = config.streams_lambda
        
        self.metrics_lambda_name = config.metrics_lambda 
        self.shared_metrics_lambda_name = f'{config.metrics_lambda}_shared'
        self.error_metrics_lambda_name = f'{config.metrics_lambda}_error'
        self.system_metrics_lambda_name = config.system_metrics_lambda

        self.experiments_lambda_name = config.experiments_lambda
        self.variants_lambda_name = f'{config.experiments_lambda}_variants'
        self.mutations_lambda_name = f'{config.experiments_lambda}_mutations'

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._client = None
        self._loop = asyncio.get_event_loop()
        self.session_uuid = str(uuid.uuid4())

        self.reporter_type = ReporterTypes.AWSLambda
        self.reporter_type_name = self.reporter_type.name.capitalize()
        self.metadata_string: str = None

        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to AWS - Region: {self.region_name}')

        self._client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'lambda',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Successfully opened connection to AWS - Region: {self.region_name}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Session System Metrics to file - {self.experiments_lambda_name}')
        
        metrics_sets: List[Dict[str, Union[int, float, str]]] = []

        for metrics_set in system_metrics_sets:

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics.record)
                
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics.record)
                
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.experiments_lambda_name,
                Payload=json.dumps(metrics_sets)
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Session System Metrics to file - {self.experiments_lambda_name}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        pass

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Streams to file - {self.streams_lambda_name}')
        
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.streams_lambda_name,
                Payload=json.dumps([
                    {
                        'stage': stream_name,
                        **stream_set.grouped 
                    } for stream_name, stream_set in stream_metrics.items()
                ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Streams to file - {self.streams_lambda_name}')


    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Experiments to file - {self.experiments_lambda_name}')
        
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.experiments_lambda_name,
                Payload=json.dumps([
                    experiment.record for experiment in experiment_metrics.experiment_summaries
                ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Experiments to file - {self.experiments_lambda_name}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Variant to file - {self.variants_lambda_name}')
        
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.variants_lambda_name,
                Payload=json.dumps([
                    variant.record for variant in experiment_metrics.variant_summaries
                ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Variant to file - {self.variants_lambda_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Mutation to file - {self.mutations_lambda_name}')
        
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.mutations_lambda_name,
                Payload=json.dumps([
                    mutation.record for mutation in experiment_metrics.mutation_summaries
                ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Mutation to file - {self.mutations_lambda_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Lambda - {self.events_lambda_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.events_lambda_name,
                Payload=json.dumps([
                    event.record for event in events
                ])
            )
        )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Lambda - {self.events_lambda_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Lambda - {self.shared_metrics_lambda_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.shared_metrics_lambda_name,
                    Payload=json.dumps([
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'common',
                            **metrics_set.common_stats
                        } for metrics_set in metrics_sets
                    ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Lambda - {self.shared_metrics_lambda_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Lambda - {self.metrics_lambda_name}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._client.invoke,
                    FunctionName=self.metrics_lambda_name,
                    Payload=json.dumps([
                        {
                            'group': group_name,
                            **group.record,
                            **group.custom
                        } for group_name, group in metrics_set.groups.items()
                    ])
                )
            )
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Lambda - {self.metrics_lambda_name}')


    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Lambda - {self.metrics_lambda_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=self.metrics_lambda_name,
                Payload=json.dumps([
                    {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'custom',
                        **{
                            metric.metric_shortname: metric.metric_value for metric in metrics_set.custom_metrics.values()
                        }
                    } for metrics_set in metrics_sets
                ])
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Lambda - {self.metrics_lambda_name}')

    async def submit_errors(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Lambda - {self.error_metrics_lambda_name}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._client.invoke,
                    FunctionName=self.error_metrics_lambda_name,
                    Payload=json.dumps([
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            **error
                        } for error in metrics_set.errors
                    ])
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Lambda - {self.error_metrics_lambda_name}')
            

    async def close(self):
        self._executor.shutdown(cancel_futures=True)
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


from .base_event import BaseEvent


class HeadersSent(BaseEvent):
    event_type='HEADERS_SENT'
    """
    The _HeadersSent event is fired whenever headers are sent.

    This is an internal event, used to determine validation steps on
    outgoing header blocks.
    """
    pass




from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictBytes
)
from typing import Union


class InjectHeaderValidator(BaseModel):
    header_name: StrictStr
    header_value: Union[StrictStr, StrictBytes, bytearray]

    class Config:
        arbitrary_types_allowed=True


import datetime

class Timer:

    def __init__(self, name: str) -> None:
        self.name = name
        self.elapsed = 0
        self.seconds = 0
        self.minutes = 0
        self.hours = 0
        self.start = datetime.datetime.now()
        self.default_message = 'Pending...'
        self.elapsed_message = self.default_message

    def __str__(self) -> str:
        return self.elapsed_message

    def update(self):
        current = datetime.datetime.now()
        self.elapsed = round((current - self.start).total_seconds())

        self.seconds = self.elapsed%60
        time_elapsed_string = f'{self.seconds}s'

        self.minutes = int(self.elapsed/60)%60
        if self.minutes > 0:
            time_elapsed_string = f'{self.minutes}m.{self.seconds}s'

        self.hours = int(self.elapsed/3600)
        if self.hours > 0:
            time_elapsed_string = f'{self.hours}h.{self.minutes}m.{self.seconds}s'


        self.elapsed_message = f'{self.name} Time Elapsed: {time_elapsed_string}'

    def finalize(self, message: str):
        self.update()
        return f'{message} - {self.elapsed_message}'

    def reset(self):
        self.start = datetime.datetime.now()
        self.elapsed_message = self.default_message